"""
Trimmed down version of aresponses.
"""

from __future__ import annotations

import asyncio
import re
from collections.abc import Awaitable, Callable, Sequence
from copy import copy
from typing import TYPE_CHECKING, Any, Self

import attrs
from aiohttp.connector import TCPConnector
from aiohttp.test_utils import RawTestServer
from aiohttp.tracing import Trace
from aiohttp.web import Response
from aiohttp.web_request import BaseRequest
from aiohttp.web_response import json_response
from yarl import URL

_Response = (
    Response
    | Callable[[BaseRequest], Response]
    | Callable[[BaseRequest], Awaitable[Response]]
    | dict[str, Any]
    | list[Any]
    | str
)


class NoRouteFoundError(AssertionError):
    pass


@attrs.frozen
class Route:
    url: URL = attrs.field(converter=URL)
    response: _Response
    method: str = attrs.field(converter=str.upper, default='GET')
    single_use: bool = False
    path_qs_pattern: re.Pattern[str] = attrs.field(init=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, 'path_qs_pattern', re.compile(self.url.path_qs, re.IGNORECASE))

    def matches(self, request: BaseRequest):
        if self.method != request.method:
            return False
        elif self.url.host != re.escape(request.host):
            return False
        elif not self.path_qs_pattern.fullmatch(request.path_qs):
            return False
        return True


class ResponsesMockServer(RawTestServer):
    def __init__(self, **kwargs: Any):
        super().__init__(handler=self.__find_response, **kwargs)
        self.__routes = list[Route]()

    async def __find_response(self, request: BaseRequest) -> Response:
        for i, route in enumerate(self.__routes):
            if not route.matches(request):
                continue

            if route.single_use:
                del self.__routes[i]

            match route.response:
                case Callable() as fn:
                    return await fn(request) if asyncio.iscoroutinefunction(fn) else fn(request)  # pyright: ignore[reportReturnType]
                case str() as text:
                    return Response(body=text)
                case (dict() | list()) as json:
                    return json_response(data=json)
                case prepared_response:
                    return copy(prepared_response)
        else:
            raise NoRouteFoundError(f'No match found for <{request.method} {request.url}>')

    def add(self, *routes: Route) -> None:
        self.__routes.extend(routes)

    @property
    def tcp_connector_class(self) -> type[TCPConnector]:
        class _TCPConnector(TCPConnector):
            async def _resolve_host(
                _self: TCPConnector, host: str, port: int, traces: Sequence[Trace] | None = None
            ) -> Any:
                return [
                    {
                        'hostname': host,
                        'host': '127.0.0.1',
                        'port': self.port,
                        'family': _self._family,
                        'proto': 0,
                        'flags': 0,
                    }
                ]

        return _TCPConnector

    if TYPE_CHECKING:
        # Parent class returns "BaseTestServer".
        async def __aenter__(self) -> Self: ...



from __future__ import annotations

import datetime as dt
import enum
from collections.abc import AsyncIterator, Callable, Collection, Iterable, Mapping, Sequence
from functools import cached_property, partial, wraps
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    NotRequired,
    Protocol,
    Self,
    TypedDict,
    overload,
)

from typing_extensions import TypeVar
from yarl import URL

from . import pkg_archives
from .catalogue import cataloguer
from .definitions import Defn, SourceMetadata
from .results import (
    AnyResult,
    PkgSourceDisabled,
    PkgSourceInvalid,
    PkgStrategiesUnsupported,
    resultify,
)

_TTokenRequired = TypeVar('_TTokenRequired', Literal[True], bool)


class AccessTokenMissingError(ValueError):
    def __str__(self) -> str:
        return 'access token missing'


class AccessToken(Generic[_TTokenRequired]):
    def __init__(self, getter: Callable[[], tuple[str | None, _TTokenRequired]]):
        self._getter = getter

    @overload
    def get(self: AccessToken[Literal[True]]) -> str: ...
    @overload
    def get(self: AccessToken[bool]) -> str | None: ...
    def get(self) -> str | None:
        access_token, required = self._getter()
        if required and access_token is None:
            raise AccessTokenMissingError
        return access_token

    @property
    def missing_reason(self) -> str | None:
        access_token, required = self._getter()
        if required and access_token is None:
            return str(AccessTokenMissingError())


class HeadersIntent(enum.IntEnum):
    Download = enum.auto()


class PkgCandidate(TypedDict):
    "A subset of the ``Pkg`` constructor's kwargs."

    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: dt.datetime
    version: str
    changelog_url: str
    deps: NotRequired[list[TypedDict[{'id': str}]]]


class Resolver(Protocol):  # pragma: no cover
    metadata: ClassVar[SourceMetadata]
    'Static source metadata.'

    access_token: ClassVar[AccessToken[bool] | None]
    'Access token retriever.'

    archive_opener: ClassVar[pkg_archives.ArchiveOpener | None]
    'Alternative archive opener to use supporting e.g. non-standard archive formats or layouts.'

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        "Attempt to extract a ``Defn`` alias from a given URL."
        ...

    def make_request_headers(self, intent: HeadersIntent | None = None) -> dict[str, str] | None:
        "Create headers for resolver HTTP requests."
        ...

    async def resolve(self, defns: Sequence[Defn]) -> dict[Defn, AnyResult[PkgCandidate]]:
        "Resolve multiple ``Defn``s into packages."
        ...

    async def resolve_one(self, defn: Defn, metadata: Any) -> PkgCandidate:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        "Retrieve a changelog from a URI."
        ...

    @classmethod
    def catalogue(cls) -> AsyncIterator[cataloguer.CatalogueEntry]:
        "Enumerate add-ons from the source."
        ...


class BaseResolver(Resolver, Protocol):
    access_token = None
    archive_opener = None

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        old_resolve_one = cls.resolve_one
        reassign_resolve_one = partial(setattr, cls, 'resolve_one')

        @reassign_resolve_one
        @wraps(old_resolve_one)
        async def _(self: Self, defn: Defn, metadata: Any):
            extraneous_strategies = defn.strategies.filled.keys() - self.metadata.strategies
            if extraneous_strategies:
                raise PkgStrategiesUnsupported(extraneous_strategies)

            return await old_resolve_one(self, defn, metadata)

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        return None

    def make_request_headers(self, intent: HeadersIntent | None = None) -> dict[str, str] | None:
        return None

    async def resolve(self, defns: Sequence[Defn]) -> dict[Defn, AnyResult[PkgCandidate]]:
        from ._utils.aio import gather
        from .progress_reporting import make_incrementing_progress_tracker

        track_progress = make_incrementing_progress_tracker(
            len(defns), f'Resolving add-ons: {self.metadata.name}'
        )
        resolve_one = resultify(self.resolve_one)
        results = await gather(track_progress(resolve_one(d, None)) for d in defns)
        return dict(zip(defns, results))

    async def get_changelog(self, uri: URL) -> str:
        match uri.scheme:
            case 'data' if uri.raw_path.startswith(','):
                import urllib.parse

                return urllib.parse.unquote(uri.raw_path[1:])

            case 'file':
                from ._utils.aio import run_in_thread
                from ._utils.web import file_uri_to_path

                return await run_in_thread(Path(file_uri_to_path(str(uri))).read_text)(
                    encoding='utf-8'
                )

            case 'http' | 'https':
                from . import http, http_ctx

                async with http_ctx.web_client().get(
                    uri,
                    expire_after=http.CACHE_INDEFINITELY,
                    headers=self.make_request_headers(),
                    raise_for_status=True,
                ) as response:
                    return await response.text()

            case _:
                raise ValueError('Unsupported URI with scheme', uri.scheme)

    @classmethod
    async def catalogue(cls) -> AsyncIterator[cataloguer.CatalogueEntry]:
        return
        yield


class Resolvers(dict[str, Resolver]):
    def __init__(self, resolvers: Iterable[type[Resolver]]):
        super().__init__((r.metadata.id, r()) for r in resolvers)

    def get_or_dummy(self, key: str) -> Resolver:
        if key in self.disabled_resolver_reasons:
            error = PkgSourceDisabled(self.disabled_resolver_reasons[key])
        elif key in self:
            return self[key]
        else:
            error = PkgSourceInvalid()

        @object.__new__
        class DummyResolver(Resolver):
            async def resolve(self, defns: Sequence[Defn]) -> dict[Defn, AnyResult[PkgCandidate]]:
                return dict.fromkeys(defns, error)

            async def get_changelog(self, uri: URL):
                raise error

            def __getattr__(self, name: str):
                raise error

        return DummyResolver

    @cached_property
    def addon_toc_key_and_id_pairs(self) -> Collection[tuple[str, str]]:
        return [
            (r.metadata.addon_toc_key, r.metadata.id)
            for r in self.values()
            if r.metadata.addon_toc_key
        ]

    @cached_property
    def archive_opener_dict(self) -> _ResolverArchiveOpenerDict:
        return _ResolverArchiveOpenerDict(self)

    @cached_property
    def disabled_resolver_reasons(self) -> Mapping[str, str]:
        return {
            r.metadata.id: d
            for r in self.values()
            if r.access_token
            for d in (r.access_token.missing_reason,)
            if d
        }

    @cached_property
    def priority_dict(self) -> _ResolverPriorityDict:
        return _ResolverPriorityDict(self)


class _ResolverArchiveOpenerDict(dict[str, pkg_archives.ArchiveOpener]):
    def __init__(self, resolvers: Resolvers) -> None:
        super().__init__(
            (r.metadata.id, r.archive_opener) for r in resolvers.values() if r.archive_opener
        )

    def __missing__(self, key: str) -> pkg_archives.ArchiveOpener:
        return pkg_archives.open_zip_archive


class _ResolverPriorityDict(dict[str, float]):
    def __init__(self, resolvers: Resolvers) -> None:
        super().__init__((n, i) for i, n in enumerate(resolvers))

    def __missing__(self, key: str) -> float:
        return float('inf')



from __future__ import annotations

import contextvars
import datetime as dt
import enum
import textwrap
from collections.abc import Awaitable, Callable, Collection, Iterable, Mapping, Sequence
from functools import partial, reduce
from itertools import chain, count, repeat
from typing import Any, Never, TypeVar, overload

import click

from .. import config as _config
from .. import config_ctx, definitions, pkg_management
from .. import results as _results
from ._helpers import ManyOptionalChoiceValueParam, SectionedHelpGroup, StrEnumChoiceParam

_T = TypeVar('_T')


class Report:
    SUCCESS_SYMBOL = click.style('✓', fg='green')
    FAILURE_SYMBOL = click.style('✗', fg='red')
    WARNING_SYMBOL = click.style('!', fg='blue')

    def __init__(
        self,
        results: Iterable[tuple[definitions.Defn, _results.Result]],
        filter_fn: Callable[[_results.Result], bool] = lambda _: True,
    ) -> None:
        self.results = list(results)
        self.filter_fn = filter_fn

    @property
    def exit_code(self) -> int:
        return any(
            isinstance(r, (_results.ManagerError, _results.InternalError)) and self.filter_fn(r)
            for _, r in self.results
        )

    @classmethod
    def _result_type_to_symbol(cls, result: _results.Result) -> str:
        match result:
            case _results.InternalError():
                return cls.WARNING_SYMBOL
            case _results.ManagerError():
                return cls.FAILURE_SYMBOL
            case _:
                return cls.SUCCESS_SYMBOL

    def __str__(self) -> str:
        return '\n'.join(
            f'{self._result_type_to_symbol(r)} {click.style(a.as_uri(), bold=True)}\n'
            f'{textwrap.fill(str(r), initial_indent=" " * 2, subsequent_indent=" " * 4)}'
            for a, r in self.results
            if self.filter_fn(r)
        )

    def generate(self) -> None:
        config = config_ctx.config()
        if config.global_config.auto_update_check:
            from .._version import is_outdated

            outdated, new_version = run_with_progress(is_outdated(config.global_config))
            if outdated:
                click.echo(f'{self.WARNING_SYMBOL} instawow v{new_version} is available')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> Never:
        self.generate()
        click.get_current_context().exit(
            self.exit_code,
        )


def run_with_progress(awaitable: Awaitable[_T]) -> _T:
    import asyncio

    from .. import http_ctx
    from ..http import init_web_client

    click_ctx = click.get_current_context().find_root()

    make_init_web_client = partial(
        init_web_client,
        config_ctx.config().global_config.http_cache_dir,
        no_cache=click_ctx.params['no_cache'],
    )

    if click_ctx.params['verbosity']:

        async def run():
            async with make_init_web_client() as web_client:
                http_ctx.web_client.set(web_client)

                return await awaitable

    else:

        async def run():
            from contextlib import AsyncExitStack

            from .._utils.aio import cancel_tasks
            from ..pkg_archives._download import PkgDownloadProgress
            from ..progress_reporting import make_progress_receiver
            from ._prompts import make_progress_bar_group

            async with AsyncExitStack() as exit_stack:
                _, make_iter_progress = exit_stack.enter_context(
                    make_progress_receiver[PkgDownloadProgress]()
                )
                update_progress = exit_stack.enter_context(make_progress_bar_group())

                async def observe_progress():
                    async for progress_group in make_iter_progress():
                        update_progress(
                            {
                                i: {
                                    'label': f'Downloading {p["defn"].as_uri()}'
                                    if p['type_'] == 'pkg_download'
                                    else p['label'],
                                    'current': p['current'],
                                    'total': p['total'],
                                    'is_download': p.get('unit') == 'bytes',
                                }
                                for i, p in progress_group.items()
                                if 'label' in p
                            }
                        )

                exit_stack.push_async_callback(
                    cancel_tasks, [asyncio.create_task(observe_progress())]
                )

                http_ctx.web_client.set(
                    await exit_stack.enter_async_context(make_init_web_client(with_progress=True))
                )

                return await awaitable
            raise

    return asyncio.run(run())


def _register_plugin_commands(group: click.Group):
    from ..plugins import get_plugin_commands

    additional_commands = (c for g in get_plugin_commands() for c in g)
    for command in additional_commands:
        group.add_command(command)
    return group


def _print_version_option(click_ctx: click.Context, _click_param: click.Parameter, value: bool):
    if not value or click_ctx.resilient_parsing:
        return

    from .._version import get_version

    click.echo(f'instawow, version {get_version()}')
    click_ctx.exit()


@overload
def _parse_defn_uri_option(
    _click_ctx: click.Context,
    _click_param: click.Parameter,
    value: str,
    *,
    retain_unknown_source: bool = False,
    raise_invalid: bool = True,
) -> definitions.Defn: ...


@overload
def _parse_defn_uri_option(
    _click_ctx: click.Context,
    _click_param: click.Parameter,
    value: list[str],
    *,
    retain_unknown_source: bool = False,
    raise_invalid: bool = True,
) -> list[definitions.Defn]: ...


def _parse_defn_uri_option(
    _click_ctx: click.Context,
    _click_param: click.Parameter,
    value: str | list[str],
    *,
    retain_unknown_source: bool = False,
    raise_invalid: bool = True,
):
    from .._utils.attrs import evolve
    from .._utils.iteration import uniq

    if not isinstance(value, str):
        defns = (
            _parse_defn_uri_option(
                _click_ctx,
                _click_param,
                v,
                retain_unknown_source=retain_unknown_source,
                raise_invalid=raise_invalid,
            )
            for v in value
        )
        return uniq(defns)

    try:
        defn = definitions.Defn.from_uri(
            value,
            known_sources=config_ctx.resolvers(),
            retain_unknown_source=retain_unknown_source,
        )
    except ValueError as exc:
        raise click.BadParameter(exc.args[0]) from None

    if not defn.source:
        match = pkg_management.get_alias_from_url(defn.alias)
        if match:
            source, alias = match
            defn = evolve(defn, {'source': source, 'alias': alias})
        elif raise_invalid and ':' not in defn.alias:
            raise click.BadParameter(value)

    return defn


def _concat_search_terms_option(
    _click_ctx: click.Context, _click_param: click.Parameter, value: tuple[str, ...]
):
    return ' '.join(value)


def _parse_iso_date_into_datetime_option(
    _click_ctx: click.Context, _click_param: click.Parameter, value: str | None
):
    if value is not None:
        return dt.datetime.strptime(value, '%Y-%m-%d').replace(tzinfo=dt.UTC)


def _make_pkg_where_clause_and_params(
    defns: Sequence[definitions.Defn],
) -> tuple[str, dict[str, Any]]:
    def make_inner():
        iter_named_param = (f'where_param_{i}' for i in count())
        for defn, source_param, alias_param in zip(defns, iter_named_param, iter_named_param):
            if not defn.source:
                yield (
                    f"(pkg.slug LIKE '%' || :{alias_param} || '%')",
                    {alias_param: defn.alias},
                )
            elif not defn.alias:
                yield (
                    f'pkg.source = :{source_param}',
                    {source_param: defn.source},
                )
            else:
                yield (
                    f'pkg.source = :{source_param} AND (pkg.id = :{alias_param} OR lower(pkg.slug) = lower(:{alias_param}))',
                    {source_param: defn.source, alias_param: defn.alias},
                )

    if defns:
        where_clauses, where_params = zip(*make_inner())
        return (
            'WHERE\n  ' + '\n  OR '.join(where_clauses),
            reduce(lambda a, b: a | b, where_params, dict[str, Any]()),
        )
    else:
        return ('', {})


class _ListFormat(enum.StrEnum):
    Simple = enum.auto()
    Detailed = enum.auto()
    Json = enum.auto()


@click.group(context_settings={'help_option_names': ('-h', '--help')}, cls=SectionedHelpGroup)
@click.option(
    '--version',
    callback=_print_version_option,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    help='Show the version and exit',
)
@click.option(
    '--verbose',
    '-v',
    'verbosity',
    count=True,
    help='Log incrementally more things.  Additive.',
)
@click.option(
    '--no-cache',
    is_flag=True,
    default=False,
    help='Disable the HTTP cache.',
)
@click.option(
    '--profile',
    '-p',
    default='__default__',
    help='Activate the specified profile.',
)
def cli(verbosity: int, no_cache: bool, profile: str):
    "Add-on manager for World of Warcraft."

    from .._logging import setup_logging

    global_config = _config.GlobalConfig.read().ensure_dirs()
    setup_logging(
        global_config.logging_dir, verbosity > 0, verbosity > 1, verbosity > 2, profile=profile
    )

    @config_ctx.config.set
    def _():
        try:
            return _config.ProfileConfig.read(global_config, profile).ensure_dirs()
        except _config.UninitialisedConfigError:
            return click.get_current_context().invoke(configure)


@cli.command
@click.argument('addons', nargs=-1, callback=_parse_defn_uri_option)
@click.option(
    '--replace',
    'replace_folders',
    is_flag=True,
    default=False,
    help='Replace unreconciled add-ons.',
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to install add-ons.  Add-on archives will not be download and the '
    'database will not be modified.',
)
def install(addons: Sequence[definitions.Defn], replace_folders: bool, dry_run: bool):
    "Install add-ons."

    results = run_with_progress(
        pkg_management.install(addons, replace_folders=replace_folders, dry_run=dry_run)
    )
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, callback=_parse_defn_uri_option)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to update add-ons.  Add-on archives will not be download and the '
    'database will not be modified.  Use this option to check for updates.',
)
def update(addons: Sequence[definitions.Defn], dry_run: bool):
    "Update installed add-ons."

    def filter_results(result: _results.Result):
        # Hide packages from output if they are up to date and not pinned.
        return True if addons or not isinstance(result, _results.PkgUpToDate) else result.is_pinned

    results = run_with_progress(pkg_management.update(addons or 'all', dry_run=dry_run))
    Report(results.items(), filter_results).generate_and_exit()


@cli.command
@click.argument(
    'addons',
    nargs=-1,
    required=True,
    callback=partial(_parse_defn_uri_option, retain_unknown_source=True),
)
@click.option(
    '--keep-folders',
    is_flag=True,
    default=False,
    help='Remove the add-on from the database but do not delete its folders.',
)
def remove(addons: Sequence[definitions.Defn], keep_folders: bool):
    "Remove add-ons."

    results = run_with_progress(pkg_management.remove(addons, keep_folders=keep_folders))
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument('addon', callback=_parse_defn_uri_option)
@click.option(
    '--undo',
    is_flag=True,
    default=False,
    help='Undo rollback by reinstalling an add-on using the default strategy.',
)
def rollback(addon: definitions.Defn, undo: bool):
    "Roll an add-on back to an older version."

    from ._prompts import Choice, select_one

    [maybe_pkg] = pkg_management.get_pinnable_pkgs([addon], for_rollback=True)
    match maybe_pkg:
        case _results.InternalError() | _results.ManagerError() as error:
            Report([(addon, error)]).generate_and_exit()

        case pkg:
            if undo:
                Report(
                    run_with_progress(
                        pkg_management.update([addon.with_default_strategy_set()])
                    ).items()
                ).generate_and_exit()

            reconstructed_defn = pkg.to_defn()

            choices = [
                Choice(v.version, value=v.version, disabled=v.version == pkg.version)
                for v in pkg.logged_versions
            ]
            selection = select_one(
                f'Select version of {reconstructed_defn.as_uri()} for rollback', choices
            ).prompt()

            Report(
                run_with_progress(
                    pkg_management.update([reconstructed_defn.with_version(selection)])
                ).items()
            ).generate_and_exit()


@cli.command
@click.option('--auto', '-a', is_flag=True, default=False, help='Do not ask for user input.')
@click.option(
    '--list-unreconciled', is_flag=True, default=False, help='List unreconciled add-ons and exit.'
)
def reconcile(auto: bool, list_unreconciled: bool):
    "Reconcile pre-installed add-ons."

    from .._utils.iteration import all_eq, uniq
    from .._utils.text import tabulate
    from ..matchers import DEFAULT_MATCHERS, AddonFolder, get_unreconciled_folders
    from ..resolvers import PkgCandidate
    from ._prompts import SKIP, Choice, confirm, select_one

    leftovers = get_unreconciled_folders()
    if list_unreconciled and leftovers:
        click.echo(tabulate([('unreconciled',), *((f.name,) for f in sorted(leftovers))]))
        return
    elif not leftovers:
        click.echo('No add-ons left to reconcile.')
        return

    if not auto:
        click.echo(
            textwrap.dedent(
                """\
                Use the arrow keys to navigate, <o> to open an add-on in your browser,
                enter to make a selection and <s> to skip to the next item.

                Versions that differ from the installed version or differ between
                choices are highlighted in purple.

                instawow will perform three passes in decreasing order of accuracy,
                looking to match source IDs and add-on names in TOC files, and folders.

                Selected add-ons will be reinstalled.

                You can also run `reconcile` in automatic mode by passing
                the `--auto` flag.  In this mode, add-ons will be reconciled
                without user input.
                """
            )
        )

    if auto:

        def confirm_install() -> bool:
            return True

        def select_pkg(
            addons: Sequence[AddonFolder], pkg_candidates: Mapping[definitions.Defn, PkgCandidate]
        ) -> definitions.Defn | None:
            return next(iter(pkg_candidates))

    else:

        def confirm_install():
            return confirm('Install selected add-ons?').prompt()

        def select_pkg(
            addons: Sequence[AddonFolder], pkg_candidates: Mapping[definitions.Defn, PkgCandidate]
        ):
            # Highlight versions if they are disparate
            highlight_version = not all_eq(
                chain(
                    (a.toc_reader.version for a in addons),
                    (p['version'] for p in pkg_candidates.values()),
                )
            )

            selection = select_one(
                f'{textwrap.shorten(", ".join(a.name for a in addons), 60)}'
                f' [{addons[0].toc_reader.version or "?"}]',
                [
                    Choice(
                        [
                            ('', f'{d.as_uri()}=='),
                            ('class:attention' if highlight_version else '', p['version']),
                        ],
                        d,
                        browser_url=p['url'],
                    )
                    for d, p in pkg_candidates.items()
                ],
                can_skip=True,
            ).prompt()
            return selection if selection is not SKIP else None

    def gather_selections(
        groups: Collection[tuple[Sequence[AddonFolder], Sequence[definitions.Defn]]],
        pkg_candidates: Mapping[definitions.Defn, PkgCandidate],
    ):
        for addon_folder, defns in groups:
            shortlist = {
                definitions.Defn(o.source, p['slug'], p['id']): p
                for o in defns
                for p in (pkg_candidates.get(o),)
                if p
            }
            if shortlist:
                selection = select_pkg(addon_folder, shortlist)
                yield selection
            else:
                yield None

    for matcher_fn in DEFAULT_MATCHERS.values():
        defn_groups = run_with_progress(matcher_fn(leftovers))
        resolve_results = run_with_progress(
            pkg_management.resolve(uniq(d for _, b in defn_groups for d in b))
        )
        pkg_candidates, _ = pkg_management.bucketise_results(resolve_results.items())

        selections = [s for s in gather_selections(defn_groups, pkg_candidates) if s]
        if selections and confirm_install():
            results = run_with_progress(pkg_management.install(selections, replace_folders=True))
            Report(results.items()).generate()

        leftovers = get_unreconciled_folders()
        if not leftovers:
            break

    if leftovers:
        click.echo()
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))


@cli.command
@click.argument('addons', nargs=-1, callback=partial(_parse_defn_uri_option, raise_invalid=False))
def rereconcile(addons: Sequence[definitions.Defn]):
    "Rereconcile installed add-ons."

    from .._utils.iteration import all_eq, uniq
    from ..pkg_db.models import Pkg, build_pkg_from_row_mapping
    from ._prompts import SKIP, Choice, confirm, select_one

    with config_ctx.database() as connection:
        query = """
            SELECT *
            FROM pkg
            {where_clause}
            ORDER BY lower(name)
        """
        if addons:
            where_clause, query_params = _make_pkg_where_clause_and_params(addons)
            execute_query = partial(
                connection.execute, query.format(where_clause=where_clause), query_params
            )
        else:
            execute_query = partial(connection.execute, query.format(where_clause=''))

        installed_pkgs = [
            build_pkg_from_row_mapping(connection, p) for p in execute_query().fetchall()
        ]

    equivalent_pkg_defn_groups = run_with_progress(
        pkg_management.find_equivalent_pkg_defns(installed_pkgs)
    )
    resolve_results = run_with_progress(
        pkg_management.resolve(uniq(d for b in equivalent_pkg_defn_groups.values() for d in b))
    )
    pkg_candidates, _ = pkg_management.bucketise_results(resolve_results.items())

    def select_alternative_pkg(pkg: Pkg, equivalent_defns: Sequence[definitions.Defn]):
        shortlisted_pkg_candidates = {
            definitions.Defn(d.source, p['slug'], p['id']): p
            for d in equivalent_defns
            for p in (pkg_candidates.get(d),)
            if p
        }
        if not shortlisted_pkg_candidates:
            return None

        highlight_version = not all_eq(
            chain((pkg.version,), (p['version'] for p in shortlisted_pkg_candidates.values()))
        )

        selection = select_one(
            pkg.name,
            [
                Choice(
                    [
                        ('', f'{d.as_uri()}=='),
                        ('class:attention' if highlight_version else '', v),
                    ],
                    d,
                    browser_url=u,
                    disabled=n,
                )
                for d, v, u, n in chain(
                    ((pkg.to_defn(), pkg.version, pkg.url, True),),
                    (
                        (d, p['version'], p['url'], False)
                        for d, p in shortlisted_pkg_candidates.items()
                    ),
                )
            ],
            can_skip=True,
        ).prompt()
        return selection if selection is not SKIP else None

    selections = {
        p.to_defn(): s
        for p, d in equivalent_pkg_defn_groups.items()
        for s in (select_alternative_pkg(p, d),)
        if s
    }

    if selections and confirm('Install selected add-ons?').prompt():
        Report(run_with_progress(pkg_management.replace(selections)).items()).generate_and_exit()


@cli.command
@click.argument('search-terms', nargs=-1, required=True, callback=_concat_search_terms_option)
@click.option(
    '--limit',
    '-l',
    default=10,
    show_default=True,
    type=click.IntRange(1, 20, clamp=True),
    help='Maximum results to return.',
)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime_option,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
@click.option(
    '--source',
    'sources',
    multiple=True,
    help='Search only in the specified source.  Multiple option.',
)
@click.option(
    '--prefer-source',
    help='Hide duplicates from other sources.',
)
@click.option(
    '--no-exclude-installed',
    is_flag=True,
    default=False,
    help='Do not exclude installed add-ons from search results.',
)
def search(
    search_terms: str,
    limit: int,
    sources: Sequence[str],
    prefer_source: str | None,
    start_date: dt.datetime | None,
    no_exclude_installed: bool,
):
    "Search for add-ons to install."

    from ..catalogue.search import search
    from ._prompts import Choice, confirm, select_multiple

    catalogue_entries = run_with_progress(
        search(
            search_terms,
            limit=limit,
            sources=frozenset(sources),
            prefer_source=prefer_source,
            start_date=start_date,
            filter_installed='ident' if no_exclude_installed else 'exclude_from_all_sources',
        )
    )
    if catalogue_entries:
        choices = [
            Choice(f'{e.name}  ({d.as_uri()})', d, browser_url=e.url)
            for e in catalogue_entries
            for d in (definitions.Defn(e.source, e.slug or e.id, e.id),)
        ]
        selections = select_multiple('Select add-ons to install', choices=choices).prompt()
        if selections:
            if confirm('Install selected add-ons?').prompt():
                click.get_current_context().invoke(
                    install, addons=selections, replace_folders=False, dry_run=False
                )
        else:
            click.echo(
                'Nothing was selected; select add-ons with <space>'
                ' and confirm by pressing <enter>.'
            )

    else:
        click.echo('No results found.')


@cli.command('list')
@click.argument('addons', nargs=-1, callback=partial(_parse_defn_uri_option, raise_invalid=False))
@click.option(
    '--format',
    '-f',
    'output_format',
    type=StrEnumChoiceParam(_ListFormat),
    default=_ListFormat.Simple,
    show_default=True,
    help='Change the output format.',
)
def list_installed(addons: Sequence[definitions.Defn], output_format: _ListFormat):
    "List installed add-ons."

    from ..pkg_db.models import Pkg, build_pkg_from_row_mapping

    with config_ctx.database() as connection:
        where_clause, where_params = _make_pkg_where_clause_and_params(addons)
        pkg_mappings = connection.execute(
            f"""
                SELECT *
                FROM pkg
                {where_clause}
                ORDER BY source, lower(name)
            """,
            where_params,
        ).fetchall()

        def row_mappings_to_pkgs():
            return map(build_pkg_from_row_mapping, repeat(connection), pkg_mappings)

        match output_format:
            case _ListFormat.Json:
                from cattrs.preconf.json import (
                    make_converter,  # pyright: ignore[reportUnknownVariableType]
                )

                click.echo(
                    make_converter().dumps(list(row_mappings_to_pkgs()), indent=2),
                )

            case _ListFormat.Detailed:
                import attrs

                def format_deps(pkg: Pkg):
                    return (
                        definitions.Defn(pkg.source, s or e.id).as_uri()
                        for e in pkg.deps
                        for (s,) in (
                            connection.execute(
                                'SELECT slug FROM pkg WHERE source = :source AND id = :id',
                                {'source': pkg.source, 'id': pkg.id},
                            ).fetchone(),
                        )
                    )

                formatter = click.HelpFormatter(max_width=99)

                for pkg in row_mappings_to_pkgs():
                    with formatter.section(pkg.to_defn().as_uri()):
                        formatter.write_dl(
                            [
                                ('name', pkg.name),
                                ('description', textwrap.shorten(pkg.description, 280)),
                                ('url', pkg.url),
                                ('version', pkg.version),
                                ('date published', pkg.date_published.astimezone().ctime()),
                                ('folders', ', '.join(f.name for f in pkg.folders)),
                                ('dependencies', ', '.join(format_deps(pkg))),
                                (
                                    'options',
                                    '; '.join(
                                        f'{s}={v!r}' for s, v in attrs.asdict(pkg.options).items()
                                    ),
                                ),
                            ]
                        )

                click.echo(formatter.getvalue(), nl=False)

            case _ListFormat.Simple:
                click.echo(
                    ''.join(
                        f'{definitions.Defn(p["source"], p["slug"]).as_uri()}\n'
                        for p in pkg_mappings
                    ),
                    nl=False,
                )


@cli.command(hidden=True)
@click.argument('addon', callback=partial(_parse_defn_uri_option, raise_invalid=False))
def info(addon: definitions.Defn):
    "Alias of `list -f detailed`."
    click.get_current_context().invoke(
        list_installed, addons=(addon,), output_format=_ListFormat.Detailed
    )


@cli.command
@click.argument('addon', callback=partial(_parse_defn_uri_option, raise_invalid=False))
def reveal(addon: definitions.Defn):
    "Bring an add-on up in your file manager."

    from .._utils.file import reveal_folder

    with config_ctx.database() as connection:
        where_clause, where_params = _make_pkg_where_clause_and_params([addon])
        pkg_folder = connection.execute(
            f"""
                SELECT pkg_folder.name
                FROM pkg
                JOIN pkg_folder ON pkg.source = pkg_folder.pkg_source AND pkg.id = pkg_folder.pkg_id
                {where_clause}
                LIMIT 1
                """,
            where_params,
        ).fetchone()

        if pkg_folder:
            reveal_folder(config_ctx.config().addon_dir / pkg_folder['name'])
        else:
            Report([(addon, _results.PkgNotInstalled())]).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, callback=partial(_parse_defn_uri_option, raise_invalid=False))
@click.option(
    '--convert/--no-convert',
    default=True,
    show_default=True,
    help='Convert HTML and Markdown changelogs to plain text using pandoc. No-op if pandoc is not installed.',
)
@click.option(
    '--remote',
    is_flag=True,
    default=False,
    help='Fetch changelogs from sources.',
)
def view_changelog(addons: Sequence[definitions.Defn], convert: bool, remote: bool):
    """View installed and remote add-on changelogs.

    Invoked without arguments, it displays the changelogs of all add-ons
    to have been installed within one minute of the last add-on.

    By default, this command will only retrieve installed add-on changelogs.
    You can reverse this behaviour by passing `--remote`.  With `--remote`,
    you are also able to retrieve changelogs of older versions from sources
    which support the `version_eq` strategy, e.g. `github:foo/bar#version_eq=v1`.
    """

    from .._utils.aio import gather
    from ..definitions import ChangelogFormat

    MAX_LINES = 100

    def make_converter():
        import shutil

        pandoc = shutil.which('pandoc')
        if pandoc is None:

            def noop_convert(source: str, changelog: str):
                return changelog

            return noop_convert
        else:
            import subprocess

            def real_convert(source: str, changelog: str):
                match config_ctx.resolvers()[source].metadata.changelog_format:
                    case ChangelogFormat.Html:
                        pandoc_input_format = 'html'
                    case ChangelogFormat.Markdown:
                        # The "markdown" format will treat a list without a preceding
                        # empty line as a paragraph, which breaks the changelog
                        # of at least one popular add-on.
                        pandoc_input_format = 'commonmark'
                    case _:
                        return changelog

                return subprocess.check_output(
                    [pandoc, '-f', pandoc_input_format, '-t', 'plain'],
                    input=changelog,
                    text=True,
                )

            return real_convert

    def format_combined_changelog_entry(source: str, slug: str, changelog: str):
        lines = changelog.splitlines()
        body = '\n'.join(
            (
                *(f'  {i}' for i in lines[:MAX_LINES]),
                *(('  [...]',) if len(lines) > MAX_LINES else ()),
            )
        )
        return f'{definitions.Defn(source, slug).as_uri()}:\n{body}'

    if remote:
        pkg_candidates, _ = pkg_management.bucketise_results(
            run_with_progress(pkg_management.resolve(addons)).items()
        )
        partial_pkgs: list[dict[str, Any]] = [
            {'source': d.source} | c for d, c in pkg_candidates.items()
        ]

    else:
        with config_ctx.database() as connection:
            query = """
                SELECT pkg.source, pkg.slug, pkg.changelog_url
                FROM pkg
            """
            if addons:
                where_clause, query_params = _make_pkg_where_clause_and_params(addons)
                execute = partial(connection.execute, query + where_clause, query_params)
            else:
                query += """
                    JOIN pkg_version_log ON pkg.source = pkg_version_log.pkg_source AND pkg.id = pkg_version_log.pkg_id
                    WHERE pkg_version_log.install_time >= (
                        SELECT datetime(max(pkg_version_log.install_time), '-1 minute')
                        FROM pkg_version_log
                        JOIN pkg ON pkg.source = pkg_version_log.pkg_source AND pkg.id = pkg_version_log.pkg_id
                    )
                """
                execute = partial(connection.execute, query)

            partial_pkgs = execute().fetchall()

    changelogs = run_with_progress(
        gather(pkg_management.get_changelog(m['source'], m['changelog_url']) for m in partial_pkgs)
    )
    if convert:
        do_convert = make_converter()
        changelogs = (do_convert(m['source'], c) for m, c in zip(partial_pkgs, changelogs))

    output = '\n\n'.join(
        format_combined_changelog_entry(m['source'], m['slug'], c)
        for m, c in zip(partial_pkgs, changelogs)
    )
    if output:
        click.echo_via_pager(output)


class _EditableConfigOptions(enum.StrEnum):
    AddonDir = 'addon_dir'
    GameFlavour = 'game_flavour'
    AutoUpdateCheck = 'global_config.auto_update_check'
    GithubAccessToken = 'global_config.access_tokens.github'
    CfcoreAccessToken = 'global_config.access_tokens.cfcore'
    WagoAddonsAccessToken = 'global_config.access_tokens.wago_addons'

    path: tuple[str, ...]

    def __new__(cls, value: str):
        self = str.__new__(cls, value)
        self._value_ = value
        self.path = tuple(value.split('.'))
        return self


@cli.command
@click.argument(
    'editable-config-values',
    nargs=-1,
    type=ManyOptionalChoiceValueParam(
        StrEnumChoiceParam(_EditableConfigOptions),
        value_types={
            _EditableConfigOptions.AutoUpdateCheck: click.BOOL,
        },
    ),
)
def configure(editable_config_values: Mapping[_EditableConfigOptions, Any]):
    """Configure instawow.

    You can pass configuration keys as arguments to reconfigure an existing
    profile.  Pass a value to suppress the interactive prompt.  For example:

    \b
    * ``configure addon_dir`` will initiate an interactive session
      with autocompletion
    * ``configure "addon_dir=~/foo"` will set ``addon_dir``'s value
      to ``~/foo`` immediately
    """

    import asyncio

    import attrs

    from .._logging import logger
    from ..wow_installations import (
        ADDON_DIR_PARTS,
        Flavour,
        find_installations,
        infer_flavour_from_addon_dir,
    )
    from ._prompts import (
        SKIP,
        Choice,
        confirm,
        make_attrs_field_validator,
        password,
        path,
        select_one,
    )

    click_ctx = click.get_current_context().find_root()
    profile = click_ctx.params['profile']

    global_config = _config.GlobalConfig.read()

    config_values: dict[str, Any] | None = None
    try:
        config_values = attrs.asdict(_config.ProfileConfig.read(global_config, profile))
    except _config.UninitialisedConfigError:
        pass
    except Exception:
        logger.exception('Unable to read existing config')

    is_new_profile = config_values is None
    if is_new_profile:
        config_values = {'profile': profile, 'global_config': attrs.asdict(global_config)}

    editable_config_values = dict(editable_config_values)
    if not editable_config_values:
        default_keys = {_EditableConfigOptions.AddonDir, _EditableConfigOptions.GameFlavour}
        if global_config.access_tokens.github is None:
            default_keys |= {_EditableConfigOptions.GithubAccessToken}

        editable_config_values = dict.fromkeys(default_keys)

    interactive_editable_config_keys = {k for k, v in editable_config_values.items() if v is None}
    if interactive_editable_config_keys:
        # 0 = Unconfigured
        # 1 = `addon_dir` configured
        # 2 = both `addon_dir` and `game_flavour` configured
        installation_configured = 0
        if (
            is_new_profile
            and {_EditableConfigOptions.AddonDir, _EditableConfigOptions.GameFlavour}
            <= interactive_editable_config_keys
        ):
            known_installations = list(global_config.iter_installations())
            unimported_installations = [
                (k, v and v['flavour'])
                for k, v in find_installations()
                if not any(d.samefile(k) for d in known_installations)
            ]
            if unimported_installations:
                selection = select_one(
                    'Installation',
                    [Choice(f'{k}  [{v}]', (k, v)) for k, v in unimported_installations],
                    can_skip=True,
                ).prompt()
                if selection is not SKIP:
                    (installation_path, flavour) = selection

                    addon_dir = installation_path.joinpath(*ADDON_DIR_PARTS)
                    addon_dir.mkdir(exist_ok=True)

                    editable_config_values |= {
                        _EditableConfigOptions.AddonDir: addon_dir,
                        _EditableConfigOptions.GameFlavour: flavour,
                    }
                    installation_configured = 2 if flavour else 1

        if (
            installation_configured == 0
            and _EditableConfigOptions.AddonDir in interactive_editable_config_keys
        ):
            editable_config_values[_EditableConfigOptions.AddonDir] = path(
                'Add-on directory',
                only_directories=True,
                validator=make_attrs_field_validator(
                    _config.ProfileConfig, 'addon_dir', _config.config_converter
                ),
            ).prompt()

        if (
            installation_configured < 2
            and _EditableConfigOptions.GameFlavour in interactive_editable_config_keys
        ):
            editable_config_values[_EditableConfigOptions.GameFlavour] = select_one(
                'Game flavour',
                [Choice(f, f) for f in Flavour if not f.is_retired],
                initial_value=config_values.get('addon_dir')
                and infer_flavour_from_addon_dir(config_values['addon_dir']),
            ).prompt()

        if _EditableConfigOptions.AutoUpdateCheck in interactive_editable_config_keys:
            editable_config_values[_EditableConfigOptions.AutoUpdateCheck] = confirm(
                'Periodically check for instawow updates?'
            ).prompt()

        if _EditableConfigOptions.GithubAccessToken in interactive_editable_config_keys:
            click.echo(
                textwrap.fill(
                    'Generating an access token for GitHub is recommended '
                    'to avoid being rate limited.  You may only perform 60 '
                    'requests an hour without an access token.'
                )
            )
            if confirm('Set up GitHub authentication?').prompt():
                from .._github_auth import get_codes, poll_for_access_token
                from ..http import init_web_client

                async def github_oauth_flow():
                    async with init_web_client(None) as web_client:
                        codes = await get_codes(web_client)
                        click.echo(
                            f'Navigate to {codes["verification_uri"]} and paste the code below:'
                        )
                        click.echo(f'  {codes["user_code"]}')
                        click.echo('Waiting...')
                        access_token = await poll_for_access_token(
                            web_client, codes['device_code'], codes['interval']
                        )
                        return access_token

                editable_config_values[_EditableConfigOptions.GithubAccessToken] = asyncio.run(
                    asyncio.wait_for(github_oauth_flow(), timeout=60 * 5)
                )

        if _EditableConfigOptions.CfcoreAccessToken in interactive_editable_config_keys:
            click.echo(
                textwrap.fill(
                    'An API key is required to use CurseForge. '
                    'Log in to CurseForge for Studios <https://console.curseforge.com/> '
                    'to generate a key.'
                )
            )
            editable_config_values[_EditableConfigOptions.CfcoreAccessToken] = (
                password('CurseForge API key:').prompt() or None
            )

        if _EditableConfigOptions.WagoAddonsAccessToken in interactive_editable_config_keys:
            click.echo(
                textwrap.fill(
                    'An access token is required to use Wago Addons. '
                    'Wago issues tokens to Patreon <https://addons.wago.io/patreon> '
                    'subscribers above a certain tier.'
                )
            )
            editable_config_values[_EditableConfigOptions.WagoAddonsAccessToken] = (
                password('Wago Addons access token:').prompt() or None
            )

    for key, value in editable_config_values.items():
        parent = config_values
        for part in key.path[:-1]:
            parent = parent[part]
        parent[key.path[-1]] = value

    config = _config.ProfileConfig.from_values(config_values)
    config.global_config.write()
    config.write()

    click.echo('Configuration written to:')
    click.echo(f'  {config.global_config.config_file}')
    click.echo(f'  {config.config_file}')

    return config


@cli.group('cache')
def _cache_group():
    "Manage the cache."


@_cache_group.command('clear')
def cache_clear():
    "Clear the instawow cache."

    import shutil

    shutil.rmtree(config_ctx.config().global_config.cache_dir)


@cli.group('debug')
def _debug_group():
    "Debug instawow."


@_debug_group.command('config')
def debug_config():
    "Print the active configuration."

    import json

    click.echo(json.dumps(config_ctx.config().unstructure_for_display(), indent=2))


@_debug_group.command('sources')
def debug_sources():
    "Print active source metadata."

    from cattrs.preconf.json import make_converter  # pyright: ignore[reportUnknownVariableType]

    json_converter = make_converter()

    click.echo(
        json_converter.dumps([r.metadata for r in config_ctx.resolvers().values()], indent=2)
    )


@_register_plugin_commands
@cli.group('plugins')
def _plugin_group():  # pyright: ignore[reportUnusedFunction]
    "Registered plug-ins."


@cli.command(hidden=True)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime_option,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
def generate_catalogue(start_date: dt.datetime | None):
    "Generate the master catalogue."

    import asyncio
    import json
    from pathlib import Path

    from ..catalogue.cataloguer import Catalogue

    catalogue = asyncio.run(Catalogue.collate(start_date))
    catalogue_json_dict = catalogue.to_json_dict()
    catalogue_path = Path(f'base-catalogue-v{catalogue.version}.json').resolve()
    catalogue_path.write_text(
        json.dumps(catalogue_json_dict, indent=2),
        encoding='utf-8',
    )
    catalogue_path.with_suffix(f'.compact{catalogue_path.suffix}').write_text(
        json.dumps(catalogue_json_dict, separators=(',', ':')),
        encoding='utf-8',
    )


main = partial(contextvars.copy_context().run, cli)



from __future__ import annotations

from collections.abc import Mapping
from enum import StrEnum
from typing import Any, Generic, TypeVar

import click
import click.types

from .._utils.iteration import bucketise

_TStrEnum = TypeVar('_TStrEnum', bound=StrEnum)


class StrEnumChoiceParam(Generic[_TStrEnum], click.Choice):
    def __init__(
        self,
        choice_enum: type[_TStrEnum],
        case_sensitive: bool = True,
    ) -> None:
        super().__init__(
            choices=list(choice_enum),
            case_sensitive=case_sensitive,
        )
        self.__choice_enum = choice_enum

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> _TStrEnum:
        converted_value = super().convert(value, param, ctx)
        return self.__choice_enum(converted_value)


class ManyOptionalChoiceValueParam(click.types.CompositeParamType):
    name = 'optional-choice-value'

    def __init__(
        self,
        choice_param: click.Choice,
        *,
        value_types: Mapping[str, click.types.ParamType] = {},
    ) -> None:
        super().__init__()
        self.__choice_param = choice_param
        self.__value_types = value_types

    def __parse_value(self, value: tuple[str, ...]):
        return (
            (k, v if s else None, self.__choice_param, vc)
            for r in value
            for k, s, v in (r.partition('='),)
            for vc in (self.__value_types.get(k),)
        )

    @property
    def arity(self) -> int:
        return -1

    def convert(
        self, value: tuple[str, ...], param: click.Parameter | None, ctx: click.Context | None
    ) -> Any:
        return {
            kc.convert(k, param, ctx): vc.convert(v, param, ctx) if vc and v else v
            for k, v, kc, vc in self.__parse_value(value)
        }

    def get_metavar(self, param: click.Parameter) -> str:
        return f'{{{",".join(self.__choice_param.choices)}}}[=VALUE]'


class SectionedHelpGroup(click.Group):
    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        command_sections = bucketise(
            ((s, c) for s, c in self.commands.items() if not c.hidden),
            key=lambda c: 'Command groups' if isinstance(c[1], click.Group) else 'Commands',
        )
        if command_sections:
            for section_name, commands in command_sections.items():
                with formatter.section(section_name):
                    limit = formatter.width - 6 - max(len(s) for s, _ in commands)
                    formatter.write_dl(
                        [(s, c.get_short_help_str(limit)) for s, c in commands],
                    )



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import Any
from unittest import mock

import prompt_toolkit.application
import prompt_toolkit.input
import prompt_toolkit.output
import pytest
from yarl import URL

import instawow._logging
import instawow.config
import instawow.config_ctx
import instawow.http
import instawow.http_ctx
from instawow.wow_installations import _DELECTABLE_DIR_NAMES, Flavour

from ._fixtures.http import ROUTES, ResponsesMockServer


def pytest_addoption(parser: pytest.Parser):
    parser.addoption('--iw-no-mock-http', action='store_true')


@pytest.fixture(autouse=True)
def anyio_backend():
    return 'asyncio'


@pytest.fixture
def caplog(caplog: pytest.LogCaptureFixture):
    handler_id = instawow._logging.logger.add(
        caplog.handler,
        format='{message}',
        level=0,
        filter=lambda record: record['level'].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    instawow._logging.logger.remove(handler_id)


@pytest.fixture
async def iw_aresponses(
    monkeypatch: pytest.MonkeyPatch,
):
    async with ResponsesMockServer() as server:
        monkeypatch.setattr('aiohttp.TCPConnector', server.tcp_connector_class)
        monkeypatch.setattr('aiohttp.ClientRequest.is_ssl', mock.Mock(return_value=False))
        yield server


@pytest.fixture(params=['foo'])
def iw_global_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    return {
        'config_dir': tmp_path / '__config__',
        'cache_dir': tmp_path / '__cache__',
        'state_dir': tmp_path / '__state__',
        'access_tokens': {
            'cfcore': request.param,
            'github': None,
            'wago_addons': request.param,
        },
    }


@pytest.fixture(params=[Flavour.Retail])
def iw_profile_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    installation_dir = (
        tmp_path
        / '__game__'
        / next(
            (k for k, v in _DELECTABLE_DIR_NAMES.items() if v['flavour'] is request.param),
            '_unknown_',
        )
    )
    addon_dir = installation_dir / 'interface' / 'addons'
    addon_dir.mkdir(parents=True)
    return {
        'profile': '__default__',
        'addon_dir': addon_dir,
        'game_flavour': request.param,
        '_installation_dir': installation_dir,
    }


@pytest.fixture
def iw_global_config(iw_global_config_values: dict[str, Any]):
    return instawow.config.GlobalConfig.from_values(iw_global_config_values).write()


@pytest.fixture
def iw_profile_config(
    iw_profile_config_values: dict[str, Any], iw_global_config: instawow.config.GlobalConfig
):
    return instawow.config.ProfileConfig.from_values(
        {'global_config': iw_global_config, **iw_profile_config_values}
    ).write()


@pytest.fixture(autouse=True)
def _iw_global_config_defaults(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
):
    for key in ('config_dir', 'cache_dir', 'state_dir'):
        monkeypatch.setenv(f'INSTAWOW_{key.upper()}', str(iw_global_config_values[key]))


@pytest.fixture
async def _iw_config_ctx(iw_profile_config: instawow.config.ProfileConfig):
    @instawow.config_ctx.config.set
    def token():
        return iw_profile_config

    yield
    instawow.config_ctx.config.reset(token)


@pytest.fixture
async def iw_web_client(
    iw_global_config: instawow.config.GlobalConfig,
):
    async with instawow.http.init_web_client(iw_global_config.http_cache_dir) as web_client:
        yield web_client


@pytest.fixture
async def _iw_web_client_ctx(iw_web_client: instawow.http.CachedSession):
    token = instawow.http_ctx.web_client.set(iw_web_client)
    yield
    instawow.http_ctx.web_client.reset(token)


@pytest.fixture(autouse=True, params=['all'])
async def _iw_mock_aiohttp_requests(
    request: pytest.FixtureRequest, iw_aresponses: ResponsesMockServer
):
    if request.config.getoption('--iw-no-mock-http') or any(
        m.name == 'iw_no_mock_http' for m in request.node.iter_markers()
    ):
        await iw_aresponses.__aexit__(*((None,) * 3))
        return

    if request.param == 'all':
        routes = ROUTES.values()
    else:
        urls = set(map(URL, request.param))
        if not urls.issubset(ROUTES.keys()):
            raise ValueError('Supplied routes must be subset of all routes')

        routes = (ROUTES[k] for k in ROUTES.keys() & urls)

    iw_aresponses.add(*routes)


@pytest.fixture(scope='module')
def _iw_mock_pt_progress_bar():
    with pytest.MonkeyPatch.context() as context:
        context.setattr('prompt_toolkit.shortcuts.progress_bar.ProgressBar', mock.MagicMock())
        yield


@pytest.fixture
async def _iw_mock_asyncio_run(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr('asyncio.run', asyncio.get_running_loop().run_until_complete)


@pytest.fixture
def iw_pt_input():
    with (
        prompt_toolkit.input.create_pipe_input() as pipe_input,
        prompt_toolkit.application.create_app_session(
            input=pipe_input, output=prompt_toolkit.output.DummyOutput()
        ),
    ):
        yield pipe_input



from __future__ import annotations

import click

from instawow import config_ctx
from instawow.cli._helpers import ManyOptionalChoiceValueParam


@click.group('weakauras-companion')
def wa_updater_command_group() -> None:
    "Manage your WeakAuras."


@wa_updater_command_group.command('build')
def build_weakauras_companion() -> None:
    "Build the WeakAuras Companion add-on."

    from instawow.cli import run_with_progress

    from ._config import PluginConfig
    from ._core import WaCompanionBuilder

    builder = WaCompanionBuilder(
        PluginConfig.read(config_ctx.config()).ensure_dirs(),
    )
    run_with_progress(builder.build())


@wa_updater_command_group.command
@click.argument(
    'collapsed-editable-config-values',
    nargs=-1,
    type=ManyOptionalChoiceValueParam(click.Choice(('access_tokens.wago',))),
)
def configure(collapsed_editable_config_values: dict[str, object]) -> None:
    "Configure the plug-in."

    from instawow.cli._prompts import password

    from ._config import PluginConfig

    wago_access_token = collapsed_editable_config_values.get('access_tokens.wago')
    if wago_access_token is None:
        wago_access_token = password('Wago access token:').prompt()
    if not wago_access_token:  # Convert to ``None`` if empty string.
        wago_access_token = None

    plugin_config = PluginConfig.from_values(
        {'access_tokens': {'wago': wago_access_token}, 'profile_config': config_ctx.config()}
    ).write()

    click.echo('Configuration written to:')
    click.echo(f'  {plugin_config.config_file}')


@wa_updater_command_group.command('list')
def list_installed_wago_auras() -> None:
    "List WeakAuras installed from Wago."
    from instawow._utils.text import tabulate

    from ._config import PluginConfig
    from ._core import WaCompanionBuilder

    builder = WaCompanionBuilder(
        PluginConfig.read(config_ctx.config()).ensure_dirs(),
    )

    installed_auras = sorted(
        (g.addon_name, a.id, a.url)
        for g in builder.extract_installed_auras()
        for v in g.auras.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('type', 'name', 'URL'), *installed_auras]))



from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import Self

import attrs

from instawow._utils.attrs import fauxfrozen
from instawow.config import ProfileConfig, SecretStr, config_converter
from instawow.config._helpers import (
    FieldMetadata,
    ensure_dirs,
    read_config,
    read_env_vars,
    write_config,
)

from . import NAME


@fauxfrozen
class _AccessTokens:
    wago: SecretStr | None = attrs.field(
        default=None,
        metadata=FieldMetadata(store=True),
    )


@fauxfrozen
class _PluginConfigStub:
    profile_config: ProfileConfig

    @property
    def cache_dir(self) -> Path:
        return self.profile_config.global_config.plugins_cache_dir / NAME

    @property
    def config_dir(self) -> Path:
        return self.profile_config.global_config.plugins_config_dir / NAME

    @property
    def profile_cache_dir(self) -> Path:
        return self.cache_dir / 'profiles' / self.profile_config.profile

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'


@fauxfrozen
class PluginConfig(_PluginConfigStub):
    access_tokens: _AccessTokens = attrs.field(
        default=_AccessTokens(),
        metadata=FieldMetadata(env=NAME, preparse_from_env=True, store=True),
    )

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, profile_config: ProfileConfig, *, env: bool = True) -> Self:
        return cls.from_values(
            {
                **read_config(cls, _PluginConfigStub(profile_config).config_file, missing_ok=True),
                'profile_config': profile_config,
            },
            env=env,
        )

    def ensure_dirs(self) -> Self:
        ensure_dirs(
            [
                self.cache_dir,
                self.config_dir,
                self.profile_cache_dir,
            ]
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file)
        return self



from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from enum import Enum, StrEnum
from functools import cache
from pathlib import Path
from typing import NotRequired, Self, TypedDict, TypeVar

from ._utils.iteration import fill

_TEnum = TypeVar('_TEnum', bound=Enum)


class _FlavourMeta(TypedDict):
    aliases: NotRequired[list[str]]
    is_retired: NotRequired[bool]


_DummyFlavourMeta = _FlavourMeta()


class Flavour(StrEnum):
    # The current Classic version is always aliased to "classic".
    # The assumption here is that should Classic not be discontinued
    # it will continue to be updated in place so that new Classic versions
    # will inherit the "_classic_" folder.  This means we won't have to
    # migrate Classic profiles either automatically or by requiring user
    # intervention for new Classic releases.
    Retail = ('retail',)
    VanillaClassic = ('vanilla_classic',)
    Classic = ('classic', _FlavourMeta(aliases=['cataclysm_classic']))
    WrathClassic = ('wrath_classic', _FlavourMeta(is_retired=True))

    is_retired: bool

    def __new__(cls, value: str, flavour_meta: _FlavourMeta = _DummyFlavourMeta) -> Self:
        self = str.__new__(cls, value)
        self._value_ = value
        for alias in flavour_meta.get('aliases', []):
            cls._value2member_map_.setdefault(alias, self)
        self.is_retired = flavour_meta.get('is_retired', False)
        return self

    @classmethod
    def from_flavour_keyed_enum(cls, flavour_keyed_enum: Enum) -> Self:
        return cls[flavour_keyed_enum.name]

    def to_flavour_keyed_enum(self, flavour_keyed_enum: type[_TEnum]) -> _TEnum:
        return flavour_keyed_enum[self.name]

    def get_flavour_groups(self, affine: bool) -> list[tuple[Flavour, ...] | None]:
        match (self, affine):
            case (self.__class__.Classic, True):
                return [(self, self.__class__.WrathClassic), None]
            case (_, True):
                return [(self,), None]
            case _:
                return [(self,)]


class FlavourVersionRange(Enum):
    Retail = (
        range(1_00_00, 1_13_00),
        range(2_00_00, 2_05_00),
        range(3_00_00, 3_04_00),
        range(4_00_00, 4_04_00),
        range(5_00_00, 12_00_00),
    )
    VanillaClassic = (range(1_13_00, 2_00_00),)
    Classic = (range(4_04_00, 5_00_00),)
    WrathClassic = (range(3_04_00, 4_00_00),)

    @classmethod
    def from_version(cls, version: int | str) -> Self | None:
        version_number = version if isinstance(version, int) else parse_version_string(version)
        return next((f for f in cls if f.contains(version_number)), None)

    def contains(self, version: int | str) -> bool:
        version_number = version if isinstance(version, int) else parse_version_string(version)
        return any(version_number in r for r in self.value)


class _Product(TypedDict):
    code: str
    flavour: Flavour


_DELECTABLE_DIR_NAMES: dict[str, _Product] = {
    '_retail_': {
        'code': 'wow',
        'flavour': Flavour.Retail,
    },
    '_ptr_': {
        'code': 'wowt',
        'flavour': Flavour.Retail,
    },
    '_xptr_': {
        'code': 'wowxptr',
        'flavour': Flavour.Retail,
    },
    '_beta_': {
        'code': 'wow_beta',
        'flavour': Flavour.Retail,
    },
    '_classic_era_': {
        'code': 'wow_classic_era',
        'flavour': Flavour.VanillaClassic,
    },
    '_classic_era_ptr_': {
        'code': 'wow_classic_era_ptr',
        'flavour': Flavour.VanillaClassic,
    },
    # '_classic_era_beta_': {
    #     'code': 'wow_classic_era_beta',
    #     'flavour': Flavour.VanillaClassic,
    # },
    '_classic_': {
        'code': 'wow_classic',
        'flavour': Flavour.Classic,
    },
    '_classic_ptr_': {
        'code': 'wow_classic_ptr',
        'flavour': Flavour.Classic,
    },
    '_classic_beta_': {
        'code': 'wow_classic_beta',
        'flavour': Flavour.Classic,
    },
}

ADDON_DIR_PARTS = ('Interface', 'AddOns')
_NORMALISED_ADDON_DIR_PARTS = tuple(map(str.casefold, ADDON_DIR_PARTS))


def _find_mac_installations():
    import subprocess

    try:
        possible_installations = subprocess.check_output(
            [
                'mdfind',
                "kMDItemCFBundleIdentifier == 'com.blizzard.worldofwarcraft'",
            ],
            text=True,
            timeout=3,
        ).splitlines()
    except subprocess.TimeoutExpired:
        pass
    else:
        for match in possible_installations:
            installation_path = Path(match).parent
            yield (installation_path, _DELECTABLE_DIR_NAMES.get(installation_path.name))


def find_installations() -> Iterator[tuple[Path, _Product | None]]:
    if sys.platform == 'darwin':
        yield from _find_mac_installations()


def _read_info(info_path: Path):
    with open(info_path, encoding='utf-8') as info:
        header = next(info)
        fields = tuple(f[: f.find('!')] for f in header.rstrip().split('|'))
        return [dict(zip(fields, e.rstrip().split('|'))) for e in info if not e.startswith('#')]


@cache
def _get_installed_versions(outer_installation_path: Path):
    return {
        e['Product']: e['Version'] for e in _read_info(outer_installation_path / '.build.info')
    }


def get_installation_dir_from_addon_dir(path_like: os.PathLike[str] | str) -> Path | None:
    path = Path(path_like)
    tail = tuple(map(str.casefold, path.parts[-3:]))
    if len(tail) == 3 and tail[1:] == _NORMALISED_ADDON_DIR_PARTS:
        return path.parents[1]


def infer_flavour_from_addon_dir(path: os.PathLike[str] | str) -> Flavour | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        try:
            return _DELECTABLE_DIR_NAMES[maybe_installation_dir.name]['flavour']
        except KeyError:
            return None


def get_installation_version_from_addon_dir(path: os.PathLike[str] | str) -> str | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        product = _DELECTABLE_DIR_NAMES.get(maybe_installation_dir.name)
        if product:
            try:
                all_versions = _get_installed_versions(maybe_installation_dir.parent)
                return all_versions[product['code']]
            except FileNotFoundError:
                return None


def parse_version_string(version_string: str) -> int:
    major, minor, patch = fill(map(int, version_string.split('.')), 0, 3)
    return major * 1_00_00 + minor * 1_00 + patch



from __future__ import annotations

from typing import Any

import pytest

from instawow import config_ctx
from instawow.definitions import ChangelogFormat, Defn, SourceMetadata, Strategies, Strategy
from instawow.resolvers import BaseResolver
from instawow.results import PkgStrategiesUnsupported

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


@pytest.mark.parametrize(
    ('iw_global_config_values', 'disabled_reason'),
    [('foo', None), (None, 'access token missing')],
    indirect=('iw_global_config_values',),
)
async def test_disabled_reason_forwarded(
    disabled_reason: str | None,
):
    assert (
        next(
            (d for d in config_ctx.resolvers().disabled_resolver_reasons.values()),
            None,
        )
        == disabled_reason
    )


async def test_unsupported_strategies_raise():
    class Resolver(BaseResolver):
        metadata = SourceMetadata(
            id='foo',
            name='Foo',
            strategies=frozenset(),
            changelog_format=ChangelogFormat.Raw,
            addon_toc_key=None,
        )

        def resolve_one(self, defn: Defn, metadata: Any):
            raise NotImplementedError

    defn = Defn(
        Resolver.metadata.id,
        'foo',
        strategies=Strategies(
            {
                Strategy.AnyFlavour: True,
                Strategy.AnyReleaseType: True,
                Strategy.VersionEq: '0',
            }
        ),
    )

    result = (await Resolver().resolve([defn]))[defn]

    assert type(result) is PkgStrategiesUnsupported
    assert (
        str(result)
        == f'strategies are not valid for source: {Strategy.AnyFlavour}, {Strategy.AnyReleaseType}, {Strategy.VersionEq}'
    )



from __future__ import annotations

import asyncio
import sys
import time
from pathlib import Path

import pytest

from instawow._utils.aio import run_in_thread
from instawow._utils.iteration import bucketise, merge_intersecting_sets
from instawow._utils.text import tabulate
from instawow._utils.web import file_uri_to_path


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


def test_tabulate_spits_out_ascii_table():
    data = [('key', 'value'), ('abc def', 'hhhdhhdhfhh'), ('dskfjsdfksdkf', 'huh')]
    assert (
        tabulate(data)
        == """\
     key          value   
-------------  -----------
abc def        hhhdhhdhfhh
dskfjsdfksdkf  huh        \
"""  # noqa: W291
    )


def test_merge_intersecting_sets_in_noncontiguous_collection():
    collection = [
        {'a'},
        {'b', 'c'},
        {'a', 'd'},
        {'e'},
        {'c', 'f'},
        {'g', 'a'},
    ]
    output = [
        {'a', 'd', 'g'},
        {'b', 'c', 'f'},
        {'e'},
    ]
    assert sorted(merge_intersecting_sets(collection)) == output


@pytest.mark.skipif(sys.platform == 'win32', reason='platform dependent')
def test_file_uri_to_path_posix_leading_slash_is_preserved():
    uri = Path('/foo/bar').as_uri()
    assert uri == 'file:///foo/bar'
    assert file_uri_to_path(uri) == '/foo/bar'


@pytest.mark.skipif(sys.platform != 'win32', reason='platform dependent')
def test_file_uri_to_path_win32_leading_slash_is_stripped():
    uri = Path('C:/foo/bar').as_uri()
    assert uri == 'file:///C:/foo/bar'
    assert file_uri_to_path(uri) == 'C:/foo/bar'


async def test_generator_in_run_in_thread_does_not_lock_up_loop():
    def foo():
        time.sleep(2)
        yield 'foo'

    async def bar():
        await asyncio.sleep(1)
        return ['bar']

    assert [
        await a
        for a in asyncio.as_completed(
            [
                run_in_thread(list)(foo()),
                bar(),
            ]
        )
    ] == [
        ['bar'],
        ['foo'],
    ]



from __future__ import annotations

import asyncio
import concurrent.futures
import contextlib
import contextvars
import os
from collections.abc import Callable, Set
from typing import TypeVar, cast

import aiohttp_client_cache
import diskcache
from typing_extensions import ParamSpec

_U = TypeVar('_U')
_P = ParamSpec('_P')


@contextlib.asynccontextmanager
async def make_cache(cache_dir: os.PathLike[str]):
    with concurrent.futures.ThreadPoolExecutor(1, '_http_cache') as executor:
        loop = asyncio.get_running_loop()

        def run_in_thread2(fn: Callable[_P, _U]):
            async def wrapper(*args: _P.args, **kwargs: _P.kwargs):
                return await loop.run_in_executor(
                    executor, lambda: contextvars.copy_context().run(fn, *args, **kwargs)
                )

            return wrapper

        # diskcache will only close the sqlite connection if it was initialised
        # in the same thread.
        cache = await run_in_thread2(diskcache.Cache)(os.fspath(cache_dir))

        class Cache(aiohttp_client_cache.BaseCache):
            @run_in_thread2
            def bulk_delete(self, keys: Set[str]):
                for key in keys:
                    cache.delete(key)

            @run_in_thread2
            def clear(self):
                cache.clear()

            @run_in_thread2
            def contains(self, key: str):
                return key in cache

            @run_in_thread2
            def delete(self, key: str):
                cache.delete(key)

            async def keys(self):
                with contextlib.suppress(StopIteration):
                    iter_keys = iter(cache)

                    while True:
                        yield (await run_in_thread2(next)(iter_keys))

            @run_in_thread2
            def read(self, key: str):
                return cache.get(key)

            @run_in_thread2
            def size(self):
                return len(cache)

            async def values(self):
                async for key in self.keys():
                    yield cast(aiohttp_client_cache.ResponseOrKey, run_in_thread2(cache.get)(key))

            @run_in_thread2
            def write(self, key: str, item: aiohttp_client_cache.ResponseOrKey):
                cache[key] = item

        try:
            yield Cache()
        finally:
            await run_in_thread2(cache.close)()



from __future__ import annotations

import importlib.resources
import json
from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import timedelta
from functools import cached_property, reduce
from itertools import chain, product
from pathlib import Path
from typing import Any, ClassVar, Literal, Self
from typing import NotRequired as N

import cattrs
from typing_extensions import TypedDict
from yarl import URL

from instawow import http, http_ctx
from instawow._logging import logger
from instawow._utils.aio import gather, run_in_thread
from instawow._utils.attrs import fauxfrozen
from instawow._utils.iteration import bucketise
from instawow._utils.perf import time_op
from instawow._utils.text import shasum
from instawow.progress_reporting import make_download_progress, make_incrementing_progress_tracker
from instawow.wow_installations import (
    get_installation_dir_from_addon_dir,
    get_installation_version_from_addon_dir,
    parse_version_string,
)

from ._config import PluginConfig

_base_url = URL('https://data.wago.io/api')

_aura_converter = cattrs.Converter()
_aura_converter.register_structure_hook(URL, lambda v, _: URL(v))
_aura_converter.register_unstructure_hook(URL, str)


class _WagoApiResponse(TypedDict):
    _id: str  # +   # Alphanumeric ID
    name: str  # +  # User-facing name
    slug: str  # +  # Slug if it has one; otherwise same as ``_id``
    url: str
    created: str  # ISO datetime
    modified: str  # ISO datetime
    game: str  # "classic" or xpac, e.g. "bfa"
    username: N[str]  # +  # Author username
    version: int  # +   # Version counter, incremented with every update
    # Semver auto-generated from ``version`` - for presentation only
    versionString: str
    changelog: _WagoApiResponse_Changelog  # +
    forkOf: N[str]  # Only present on forks
    regionType: N[str]  # Only present on WAs


class _WagoApiResponse_Changelog(TypedDict):
    format: N[Literal['bbcode', 'markdown']]
    text: N[str]


@fauxfrozen(kw_only=True)
class Aura:
    id: str
    uid: str
    parent: str | None = None
    url: URL
    version: int


@fauxfrozen
class AuraGroup:
    api_ep: ClassVar[str]
    addon_name: ClassVar[str]

    auras: dict[str, list[Aura]]

    @classmethod
    def from_lua_table(cls, lua_table: Mapping[str, Any]) -> Self:
        raise NotImplementedError


@fauxfrozen
class WeakAuras(AuraGroup):
    api_ep = 'weakauras'
    addon_name = 'WeakAuras'

    @classmethod
    def from_lua_table(cls, lua_table: Mapping[str, Any]):
        raw_auras = lua_table['WeakAurasSaved']['displays']
        prepared_auras = (t for t in raw_auras.values() if t.get('url'))
        auras = _aura_converter.structure(prepared_auras, list[Aura])
        sorted_auras = sorted(filter(lambda a: a.url.host == 'wago.io', auras), key=lambda a: a.id)
        return cls(bucketise(sorted_auras, key=lambda a: a.url.parts[1]))


@fauxfrozen
class Plateroos(AuraGroup):
    api_ep = 'plater'
    addon_name = 'Plater'

    @classmethod
    def from_lua_table(cls, lua_table: Mapping[str, Any]):
        raw_auras = lua_table['PlaterDB']['profiles']
        prepared_auras = (
            t | {'id': t['Name'], 'uid': ''}
            for n, p in raw_auras.items()
            for t in chain(
                [{**p, 'Name': f'__profile_{n}__'}],
                p.get('script_data') or [],
                p.get('hook_data') or [],
            )
            if t.get('url')
        )
        auras = _aura_converter.structure(prepared_auras, list[Aura])
        sorted_auras = sorted(filter(lambda a: a.url.host == 'wago.io', auras), key=lambda a: a.id)
        return cls({a.url.parts[1]: [a] for a in sorted_auras})


@fauxfrozen
class _BuildPaths:
    archive: Path
    changelog: Path
    version: Path


def _extract_auras(model: type[AuraGroup], source: str):
    from ._custom_slpp import loads

    return model.from_lua_table(loads(f'{{{source}}}'))


class WaCompanionBuilder:
    def __init__(self, config: PluginConfig) -> None:
        self._config = config

    @cached_property
    def build_paths(self) -> _BuildPaths:
        return _BuildPaths(
            self._config.profile_cache_dir / 'WeakAurasCompanion.zip',
            self._config.profile_cache_dir / 'CHANGELOG.md',
            self._config.profile_cache_dir / 'version.txt',
        )

    def _make_request_headers(self):
        access_token = self._config.access_tokens.wago
        if access_token:
            return {'api-key': access_token}

    async def _fetch_wago_metadata(self, aura_group: AuraGroup):
        request = http_ctx.web_client().post(
            (_base_url / 'check' / aura_group.api_ep),
            expire_after=timedelta(minutes=30),
            headers=self._make_request_headers(),
            json={'ids': list(aura_group.auras)},
            trace_request_ctx={'progress': make_download_progress(label='Fetching aura metadata')},
        )
        async with request as response:
            metadata: list[_WagoApiResponse]
            if response.status == 404:
                metadata = []
                return metadata

            response.raise_for_status()

            metadata = await response.json()
            return sorted(metadata, key=lambda r: r['slug'])

    async def _fetch_wago_import_string(self, remote_aura: _WagoApiResponse):
        request = http_ctx.web_client().get(
            (_base_url / 'raw' / 'encoded')
            .with_query(id=remote_aura['_id'])
            .with_fragment(str(remote_aura['version'])),
            expire_after=http.CACHE_INDEFINITELY,
            headers=self._make_request_headers(),
            raise_for_status=True,
        )
        async with request as response:
            return await response.text()

    async def _get_remote_auras(
        self, aura_group: AuraGroup
    ) -> list[tuple[Sequence[Aura], _WagoApiResponse, str]]:
        if not aura_group.auras:
            return []

        metadata = await self._fetch_wago_metadata(aura_group)

        track_progress = make_incrementing_progress_tracker(
            len(metadata), f'Fetching import strings: {aura_group.addon_name}'
        )

        import_strings = await gather(
            track_progress(self._fetch_wago_import_string(r)) for r in metadata
        )
        return [
            (aura_group.auras.get(r['slug']) or aura_group.auras[r['_id']], r, i)
            for r, i in zip(metadata, import_strings)
        ]

    def _generate_addon(
        self,
        auras: Iterable[
            tuple[type[AuraGroup], list[tuple[Sequence[Aura], _WagoApiResponse, str]]]
        ],
    ):
        from zipfile import ZipFile, ZipInfo

        from . import _templates

        template_resources = importlib.resources.files(_templates)

        aura_dict = dict.fromkeys((WeakAuras, Plateroos), ()) | dict(auras)

        with ZipFile(self.build_paths.archive, 'w') as file:

            def write_file(filename: str, content: str):
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp would be set to the current time
                # which would render the build unreproducible.
                file.writestr(
                    ZipInfo(filename=f'WeakAurasCompanion/{filename}'),
                    content,
                )
                return content

            def make_slug_entry(metadata: _WagoApiResponse, import_string: str):
                return f"""\
            [ [=[{metadata['slug']}]=] ] = {{
                name = [=[{metadata['name']}]=],
                author = [=[{metadata.get('username', '__unknown__')}]=],
                encoded = [=[{import_string}]=],
                wagoVersion = [=[{metadata['version']}]=],
                wagoSemver = [=[{metadata['version']}]=],
                versionNote = [=[{metadata['changelog'].get('text', '')}]=],
                source = [=[Wago]=],
            }},"""

            NL = '\n'
            data_output = write_file(
                'data.lua',
                f"""\
-- file generated automatically
WeakAurasCompanionData = {{
{
                    NL.join(
                        f'''
    {c.addon_name} = {{
        slugs = {{
{NL.join(make_slug_entry(m, i) for _, m, i in v)}
        }},
        stash = {{
        }},
        stopmotionFiles = {{
        }},
    }},'''
                        for c, v in aura_dict.items()
                    )
                }
}}
""",
            )

            init_output = write_file(
                'init.lua', template_resources.joinpath('init.lua.tpl').read_text()
            )

            interface_version_string = get_installation_version_from_addon_dir(
                self._config.profile_config.addon_dir
            )
            interface_version = (
                parse_version_string(interface_version_string) if interface_version_string else 0
            )
            addon_version = shasum(data_output, init_output, interface_version)[:7]

            toc_tpl = template_resources.joinpath('WeakAurasCompanion.toc.tpl').read_text()
            write_file(
                'WeakAurasCompanion.toc',
                toc_tpl.format(interface=interface_version, version=addon_version),
            )

        changelog_tpl = template_resources.joinpath('CHANGELOG.md.tpl').read_text()
        self.build_paths.changelog.write_text(
            '\n\n'.join(
                changelog_tpl.format(
                    name=a.id,
                    url=a.url.parent,
                    version=metadata['version'],
                    changelog=metadata['changelog'].get('text') or 'n/a',
                )
                for v in aura_dict.values()
                for existing_auras, metadata, _ in v
                for a in (next((i for i in existing_auras if not i.parent), existing_auras[0]),)
                if a.version != metadata['version']
            )
            or 'n/a',
            encoding='utf-8',
        )

        self.build_paths.version.write_text(
            addon_version,
            encoding='utf-8',
        )

    def extract_installed_auras(self) -> Iterator[AuraGroup]:
        installation_dir = get_installation_dir_from_addon_dir(
            self._config.profile_config.addon_dir
        )
        if not installation_dir:
            raise ValueError(
                f'cannot extract installation folder from {self._config.profile_config.addon_dir}'
            )

        saved_vars_by_account = installation_dir.glob('WTF/Account/*/SavedVariables')
        for saved_vars, model in product(
            saved_vars_by_account,
            [WeakAuras, Plateroos],
        ):
            file = (saved_vars / model.addon_name).with_suffix('.lua')
            if not file.exists():
                logger.info(f'{file} not found')

            else:
                content = file.read_text(encoding='utf-8-sig', errors='replace')
                aura_group_cache = (self._config.cache_dir / shasum(content)).with_suffix('.json')
                if aura_group_cache.exists():
                    logger.info(f'Loading {file} from cache at {aura_group_cache}')
                    aura_group_json = json.loads(aura_group_cache.read_bytes())
                    aura_group = _aura_converter.structure({'auras': aura_group_json}, model)
                else:
                    with time_op(
                        lambda t: logger.debug(
                            f'Extracted {model.__name__} in {t:.3f}s'  # noqa: B023
                        )
                    ):
                        aura_group = _extract_auras(model, content)
                    aura_group_cache.write_text(
                        json.dumps(_aura_converter.unstructure(aura_group.auras)),
                        encoding='utf-8',
                    )

                yield aura_group

    async def build(self) -> None:
        installed_auras = await run_in_thread(list)(self.extract_installed_auras())
        installed_auras_by_type = {
            t: t(reduce(lambda a, b: a | b, (i.auras for i in a)))
            for t, a in bucketise(installed_auras, key=type).items()
        }
        remote_auras = await gather(
            self._get_remote_auras(r) for r in installed_auras_by_type.values()
        )
        await run_in_thread(self._generate_addon)(zip(installed_auras_by_type, remote_auras))



from __future__ import annotations

import re
from collections.abc import Awaitable, Iterable, Mapping
from itertools import chain, product
from pathlib import Path
from typing import Protocol, Self

import attrs

from .. import config_ctx
from .._utils.attrs import fauxfrozen
from .._utils.iteration import bucketise, merge_intersecting_sets, uniq
from ..catalogue import synchronise as synchronise_catalogue
from ..definitions import Defn
from ..wow_installations import Flavour
from .addon_toc import TocReader


class Matcher(Protocol):  # pragma: no cover
    def __call__(
        self, leftovers: frozenset[AddonFolder]
    ) -> Awaitable[list[tuple[list[AddonFolder], list[Defn]]]]: ...


# https://github.com/Stanzilla/WoWUIBugs/issues/68#issuecomment-830351390
# https://warcraft.wiki.gg/wiki/TOC_format#Multiple_client_flavors
FLAVOUR_TOC_IDS = {
    Flavour.Retail: ('Mainline',),
    Flavour.VanillaClassic: ('Vanilla', 'Classic'),
    Flavour.Classic: ('Cata', 'Classic'),
    Flavour.WrathClassic: ('Wrath', 'WOTLKC', 'Classic'),
}

FLAVOUR_TOC_SUFFIXES = {
    k: tuple(f'{s}{f}.toc' for s, f in product('-_', v)) for k, v in FLAVOUR_TOC_IDS.items()
}
NORMALISED_FLAVOUR_TOC_SUFFIXES = {
    k: tuple(i.lower() for i in v) for k, v in FLAVOUR_TOC_SUFFIXES.items()
}


@fauxfrozen(order=True)
class AddonFolder:
    path: Path = attrs.field(eq=False, order=False)
    toc_reader: TocReader = attrs.field(eq=False, order=False)
    name: str = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        object.__setattr__(self, 'name', self.path.name)

    @classmethod
    def from_path(cls, flavour: Flavour, path: Path) -> Self | None:
        for suffix in chain(FLAVOUR_TOC_SUFFIXES[flavour], ('.toc',)):
            try:
                toc_reader = TocReader.from_path(path / (path.name + suffix))
                return cls(path, toc_reader)
            except FileNotFoundError:
                pass

    def get_defns_from_toc_keys(self, keys_and_ids: Iterable[tuple[str, str]]) -> frozenset[Defn]:
        return frozenset(
            Defn(s, i) for k, s in keys_and_ids for i in (self.toc_reader.get(k),) if i
        )


def _get_unreconciled_folders():
    config = config_ctx.config()

    with config_ctx.database() as connection:
        pkg_folders = [n for (n,) in connection.execute('SELECT name FROM pkg_folder').fetchall()]

    unreconciled_folder_paths = (
        p
        for p in config.addon_dir.iterdir()
        if p.name not in pkg_folders and p.is_dir() and not p.is_symlink()
    )
    for path in unreconciled_folder_paths:
        addon_folder = AddonFolder.from_path(config.game_flavour, path)
        if addon_folder:
            yield addon_folder


def get_unreconciled_folders() -> frozenset[AddonFolder]:
    return frozenset(_get_unreconciled_folders())


async def _match_toc_source_ids(leftovers: frozenset[AddonFolder]):
    resolvers = config_ctx.resolvers()
    catalogue = await synchronise_catalogue()

    def get_catalogue_defns(extracted_defns: Iterable[Defn]):
        for defn in extracted_defns:
            entry = catalogue.keyed_entries.get((defn.source, defn.alias))
            if entry:
                for addon_key in entry.same_as:
                    if addon_key.source in resolvers:
                        yield Defn(addon_key.source, addon_key.id)

    def get_addon_and_defn_pairs():
        for addon in sorted(leftovers):
            defns = addon.get_defns_from_toc_keys(resolvers.addon_toc_key_and_id_pairs)
            if defns:
                yield (addon, defns | frozenset(get_catalogue_defns(defns)))

    matches = list(get_addon_and_defn_pairs())
    merged_defns_by_constituent_defn = {
        i: s for s in merge_intersecting_sets(d for _, d in matches) for i in s
    }
    folders_grouped_by_overlapping_defns = bucketise(
        matches,
        lambda i: merged_defns_by_constituent_defn[next(iter(i[1]))],
    )
    return [
        (
            [n for n, _ in f],
            sorted(b, key=lambda d: resolvers.priority_dict[d.source]),
        )
        for b, f in folders_grouped_by_overlapping_defns.items()
    ]


async def _match_folder_name_subsets(leftovers: frozenset[AddonFolder]):
    config = config_ctx.config()
    resolvers = config_ctx.resolvers()
    catalogue = await synchronise_catalogue()

    leftovers_by_name = {e.name: e for e in leftovers}

    matches = [
        (frozenset(leftovers_by_name[n] for n in m), Defn(i.source, i.id))
        for i in catalogue.entries
        if config.game_flavour in i.game_flavours
        for f in i.folders
        for m in (f & leftovers_by_name.keys(),)
        if m
    ]

    merged_folders_by_constituent_folder = {
        i: s for s in merge_intersecting_sets(f for f, _ in matches) for i in s
    }
    matches_grouped_by_overlapping_folder_names = bucketise(
        matches, lambda v: merged_folders_by_constituent_folder[next(iter(v[0]))]
    )

    def sort_key(folders: frozenset[AddonFolder], defn: Defn):
        return (-len(folders), resolvers.priority_dict[defn.source])

    return [
        (sorted(f), uniq(d for _, d in sorted(b, key=lambda v: sort_key(*v))))
        for f, b in matches_grouped_by_overlapping_folder_names.items()
    ]


async def _match_addon_names_with_folder_names(leftovers: frozenset[AddonFolder]):
    def normalise(value: str):
        return re.sub(r'[^0-9A-Za-z]', '', value.casefold())

    catalogue = await synchronise_catalogue()

    addon_names_to_catalogue_entries = bucketise(
        catalogue.entries, key=lambda i: normalise(i.name)
    )
    matches = (
        (a, addon_names_to_catalogue_entries.get(normalise(a.name))) for a in sorted(leftovers)
    )
    return [([a], uniq(Defn(i.source, i.id) for i in m)) for a, m in matches if m]


# In order of increasing heuristicitivenessitude
DEFAULT_MATCHERS: Mapping[str, Matcher] = {
    'toc_source_ids': _match_toc_source_ids,
    'folder_name_subsets': _match_folder_name_subsets,
    'addon_names_with_folder_names': _match_addon_names_with_folder_names,
}



from __future__ import annotations

from contextlib import asynccontextmanager
from contextvars import ContextVar
from functools import partial
from pathlib import Path
from shutil import move
from tempfile import NamedTemporaryFile
from typing import Literal

from .. import config_ctx, http, http_ctx, sync_ctx
from .._utils.aio import run_in_thread
from .._utils.text import shasum
from .._utils.web import file_uri_to_path, is_file_uri
from ..definitions import Defn
from ..progress_reporting import Progress
from ..resolvers import HeadersIntent


class PkgDownloadProgress(Progress[Literal['pkg_download'], Literal['bytes']]):
    profile: str
    defn: Defn


_DOWNLOAD_PKG_LOCK = '_DOWNLOAD_PKG_'

_AsyncNamedTemporaryFile = run_in_thread(NamedTemporaryFile)
_move_async = run_in_thread(move)

_use_alt_ssl_context = ContextVar('_use_alt_ssl_context', default=False)
_alt_ssl_context = http.get_ssl_context(cloudflare_compat=True)


@asynccontextmanager
async def _open_temp_writer_async():
    fh = await _AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, run_in_thread(fh.write))
    except BaseException:
        await run_in_thread(fh.close)()
        await run_in_thread(path.unlink)()
        raise
    else:
        await run_in_thread(fh.close)()


async def download_pkg_archive(defn: Defn, download_url: str) -> Path:
    if is_file_uri(download_url):
        return Path(file_uri_to_path(download_url))

    async with sync_ctx.locks()[_DOWNLOAD_PKG_LOCK, download_url]:
        headers = config_ctx.resolvers()[defn.source].make_request_headers(
            intent=HeadersIntent.Download
        )
        trace_request_ctx = {
            'progress': PkgDownloadProgress(
                type_='pkg_download',
                unit='bytes',
                current=0,
                total=0,
                profile=config_ctx.config().profile,
                defn=defn,
            )
        }

        make_request = partial(
            http_ctx.web_client().get,
            download_url,
            headers=headers,
            trace_request_ctx=trace_request_ctx,
            expire_after=http.CACHE_INDEFINITELY,
        )
        if _use_alt_ssl_context.get():
            make_request = partial(make_request, ssl=_alt_ssl_context)

        async with (
            make_request() as response,
            _open_temp_writer_async() as (temp_path, write),
        ):
            # CloudFlare seems to think the default combination of TLS 1.2 and 1.3
            # is unacceptable: https://github.com/aio-libs/aiohttp/discussions/9300#discussioncomment-10775829
            if (
                not _use_alt_ssl_context.get()
                and response.status == 403
                and 'CF-RAY' in response.headers
            ):
                _use_alt_ssl_context.set(True)
                return await download_pkg_archive(defn, download_url)

            response.raise_for_status()

            async for chunk, _ in response.content.iter_chunks():
                await write(chunk)

        return await _move_async(
            temp_path,
            config_ctx.config().global_config.install_cache_dir / shasum(download_url),
        )



from __future__ import annotations

from datetime import datetime, timedelta
from enum import StrEnum
from typing import Never

from typing_extensions import TypedDict
from yarl import URL

from .. import config_ctx, http_ctx
from .._utils.web import as_plain_text_data_url
from ..definitions import ChangelogFormat, Defn, SourceMetadata, Strategy
from ..resolvers import AccessToken, BaseResolver, HeadersIntent, PkgCandidate
from ..results import PkgFilesNotMatching, PkgNonexistent


class _WagoStability(StrEnum):
    "https://addons.wago.io/api/data/game"

    Stable = 'stable'
    Beta = 'beta'
    Alpha = 'alpha'


class _WagoGameVersion(StrEnum):
    Retail = 'retail'
    VanillaClassic = 'classic'
    Classic = 'cata'
    WrathClassic = 'wotlk'


class _WagoAddon(TypedDict):
    "``/addons/{id}``"

    id: str
    slug: str
    display_name: str  # Eq to ``WagoRecentAddon.name``
    thumbnail_image: str | None  # Eq to ``WagoRecentAddon.thumbnail``
    summary: str
    description: str  # Long description
    website: str  # Author website
    gallery: list[str]
    authors: list[str]
    download_count: int
    website_url: str  # Page on Wago
    recent_release: dict[_WagoStability, _WagoAddonRelease] | list[Never]


class _WagoAddonRelease(TypedDict):
    label: str  # Version
    supported_retail_patch: str | None  # e.g. "9.2.5"
    supported_classic_patch: str | None
    supported_wotlk_patch: str | None
    changelog: str
    stability: _WagoStability
    created_at: str  # ISO date-time
    download_link: str


class WagoResolver(BaseResolver):
    metadata = SourceMetadata(
        id='wago',
        name='Wago Addons',
        strategies=frozenset({Strategy.AnyReleaseType}),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key='X-Wago-ID',
    )

    __wago_external_api_url = URL('https://addons.wago.io/api/external')

    @AccessToken
    def access_token():
        return config_ctx.config().global_config.access_tokens.wago_addons, True

    @classmethod
    def get_alias_from_url(cls, url: URL):
        if url.host == 'addons.wago.io' and len(url.parts) > 2 and url.parts[1] == 'addons':
            return url.parts[2]

    def make_request_headers(self, intent: HeadersIntent | None = None):
        return {'Authorization': f'Bearer {self.access_token.get()}'}

    async def resolve_one(self, defn: Defn, metadata: None):
        async with http_ctx.web_client().get(
            (self.__wago_external_api_url / 'addons' / defn.alias).with_query(
                game_version=config_ctx.config().game_flavour.to_flavour_keyed_enum(
                    _WagoGameVersion
                )
            ),
            expire_after=timedelta(minutes=5),
            headers=self.make_request_headers(),
        ) as response:
            if response.status == 404:
                raise PkgNonexistent
            response.raise_for_status()

            addon_metadata: _WagoAddon = await response.json()

        recent_releases = dict(addon_metadata['recent_release'])
        if not defn.strategies[Strategy.AnyReleaseType] and (
            stable_release := recent_releases.get(_WagoStability.Stable)
        ):
            recent_releases = {_WagoStability.Stable: stable_release}

        try:
            file_date, file = max(
                (datetime.fromisoformat(f['created_at']), f) for f in recent_releases.values()
            )
        except ValueError:
            raise PkgFilesNotMatching(defn.strategies)

        return PkgCandidate(
            id=addon_metadata['id'],
            slug=addon_metadata['slug'],
            name=addon_metadata['display_name'],
            description=addon_metadata['summary'],
            url=addon_metadata['website_url'],
            download_url=file['download_link'],
            date_published=file_date,
            version=file['label'],
            changelog_url=as_plain_text_data_url(file['changelog']),
        )



from __future__ import annotations

import contextvars as cv
from collections.abc import Mapping
from contextlib import AbstractAsyncContextManager
from typing import TypeAlias

Locks: TypeAlias = Mapping[object, AbstractAsyncContextManager[None]]


@object.__new__
class _DummyLock(AbstractAsyncContextManager[None]):
    async def __aexit__(self, *args: object):
        pass


class _dummy_lock_dict(dict[object, _DummyLock]):
    def __missing__(self, key: object):
        return _DummyLock


_locks_var = cv.ContextVar[Locks]('_locks_var', default=_dummy_lock_dict())  # noqa: B039


@object.__new__
class locks:
    def __call__(self) -> Locks:
        return _locks_var.get()

    set = staticmethod(_locks_var.set)
    reset = staticmethod(_locks_var.reset)



from __future__ import annotations

import contextvars as cv

from . import http

_web_client_var = cv.ContextVar['http.CachedSession']('_web_client_var')


@object.__new__
class web_client:
    def __call__(self) -> http.CachedSession:
        return _web_client_var.get()

    set = staticmethod(_web_client_var.set)
    reset = staticmethod(_web_client_var.reset)



from __future__ import annotations

from typing import TypedDict

from . import NAME, config


class _SimpleApiProject(TypedDict):
    versions: list[str]


def get_version() -> str:
    import importlib.metadata

    try:
        return importlib.metadata.version(NAME)
    except importlib.metadata.PackageNotFoundError:
        return '0+dev'


async def is_outdated(global_config: config.GlobalConfig) -> tuple[bool, str]:
    """Check on PyPI to see if instawow is outdated.

    The response is cached for 24 hours.
    """
    from datetime import timedelta

    from aiohttp import ClientError
    from packaging.version import Version

    from .http import init_web_client

    if not global_config.auto_update_check:
        return (False, '')

    __version__ = get_version()
    parsed_version = Version(__version__)
    if parsed_version.local:
        return (False, '')

    try:
        async with (
            init_web_client(
                global_config.http_cache_dir,
                raise_for_status=True,
            ) as web_client,
            web_client.get(
                'https://pypi.org/simple/instawow',
                expire_after=timedelta(days=1),
                headers={
                    'Accept': 'application/vnd.pypi.simple.v1+json',
                },
            ) as response,
        ):
            metadata: _SimpleApiProject = await response.json()
            version = max(metadata['versions'], key=Version)
    except ClientError:
        version = __version__

    return (Version(version) > parsed_version, version)



from __future__ import annotations

import os
import sys
from collections.abc import Iterator, Mapping, Sized
from functools import lru_cache
from pathlib import Path
from typing import NewType, Self, TypeVar

import attrs

from .. import NAME
from .._utils.attrs import enrich_validator_exc, fauxfrozen
from .._utils.file import trash
from ..wow_installations import Flavour, get_installation_dir_from_addon_dir
from ._helpers import (
    FieldMetadata,
    ensure_dirs,
    make_config_converter,
    read_config,
    read_env_vars,
    write_config,
)
from ._helpers import UninitialisedConfigError as UninitialisedConfigError
from ._helpers import config_converter as config_converter

_T = TypeVar('_T')
_TSized = TypeVar('_TSized', bound=Sized)


SecretStr = NewType('SecretStr', str)


def _expand_path(value: os.PathLike[str]):
    return Path(value).expanduser().resolve()


def _is_writable_dir(value: Path):
    return value.is_dir() and os.access(value, os.W_OK)


@enrich_validator_exc
def _validate_path_is_writable_dir(_model: object, _attr: attrs.Attribute[Path], value: Path):
    if not _is_writable_dir(value):
        raise ValueError(f'"{value}" is not a writable directory')


def _make_validate_min_length(min_length: int):
    @enrich_validator_exc
    def _validate_min_length(_model: object, _attr: attrs.Attribute[_TSized], value: _TSized):
        if len(value) < min_length:
            raise ValueError(f'Value must have a minimum length of {min_length}')

    return _validate_min_length


def _unstructure_config_for_display(config: object):
    converter = _make_display_converter()
    return converter.unstructure(config)


@lru_cache(1)
def _make_display_converter():
    def convert_secret_string(value: str):
        return '**********'

    def convert_global_config(global_config: GlobalConfig):
        return converter.unstructure_attrs_asdict(global_config) | {
            'profiles': list(global_config.iter_profiles())
        }

    converter = make_config_converter()
    converter.register_unstructure_hook(GlobalConfig, convert_global_config)
    converter.register_unstructure_hook(SecretStr, convert_secret_string)
    return converter


def _get_default_config_dir():
    parent_dir = os.environ.get('XDG_CONFIG_HOME')

    if not parent_dir:
        if sys.platform == 'darwin':
            parent_dir = Path.home() / 'Library' / 'Application Support'
        elif sys.platform == 'win32':
            parent_dir = os.environ.get('APPDATA')

    if not parent_dir:
        parent_dir = Path.home() / '.config'

    return Path(parent_dir, NAME)


def _get_default_cache_dir():
    parent_dir = os.environ.get('XDG_CACHE_HOME')

    if not parent_dir:
        if sys.platform == 'darwin':
            parent_dir = Path.home() / 'Library' / 'Caches'
        elif sys.platform == 'win32':
            parent_dir = os.environ.get('LOCALAPPDATA')

    if not parent_dir:
        parent_dir = Path.home() / '.cache'

    return Path(parent_dir, NAME)


def _get_default_state_dir():
    parent_dir = os.environ.get('XDG_STATE_HOME')

    if not parent_dir and sys.platform not in {'darwin', 'win32'}:
        parent_dir = Path.home() / '.local' / 'state'

    if not parent_dir:
        return _get_default_config_dir()

    return Path(parent_dir, NAME)


@fauxfrozen
class _AccessTokens:
    cfcore: SecretStr | None = attrs.field(
        default=None,
        metadata=FieldMetadata(store=True),
    )
    github: SecretStr | None = attrs.field(
        default=None,
        metadata=FieldMetadata(store=True),
    )
    wago_addons: SecretStr | None = attrs.field(
        default=None,
        metadata=FieldMetadata(store=True),
    )


@fauxfrozen
class GlobalConfig:
    config_dir: Path = attrs.field(
        factory=_get_default_config_dir,
        converter=_expand_path,
        metadata=FieldMetadata(env=True),
    )
    cache_dir: Path = attrs.field(
        factory=_get_default_cache_dir,
        converter=_expand_path,
        metadata=FieldMetadata(env=True),
    )
    state_dir: Path = attrs.field(
        factory=_get_default_state_dir,
        converter=_expand_path,
        metadata=FieldMetadata(env=True),
    )
    auto_update_check: bool = attrs.field(
        default=True,
        metadata=FieldMetadata(env=True, preparse_from_env=True, store=True),
    )
    access_tokens: _AccessTokens = attrs.field(
        default=_AccessTokens(),
        metadata=FieldMetadata(env=True, preparse_from_env=True, store=True),
    )

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, env: bool = True) -> Self:
        unsaved_config = cls.from_values(env=env)
        config_values = read_config(cls, unsaved_config.config_file, missing_ok=True)
        return cls.from_values(config_values, env=env) if config_values else unsaved_config

    def iter_profiles(self) -> Iterator[str]:
        "Get the names of the profiles contained in ``config_dir``."
        yield from (c.parent.name for c in self.profiles_config_dir.glob('*/config.json'))

    def iter_installations(self) -> Iterator[Path]:
        """Get the installation directory of each profile, where one can be extracted.

        *instawow* does not mandate that an add-on directory is tied to an installation
        directory which allows to e.g. update add-ons out of band.
        """
        for config_json in self.profiles_config_dir.glob('*/config.json'):
            addon_dir = read_config(ProfileConfig, config_json)['addon_dir']
            maybe_installation_dir = get_installation_dir_from_addon_dir(addon_dir)
            if maybe_installation_dir:
                yield maybe_installation_dir

    def ensure_dirs(self) -> Self:
        ensure_dirs(
            [
                self.config_dir,
                self.cache_dir,
                self.state_dir,
                self.logging_dir,
                self.http_cache_dir,
                self.install_cache_dir,
            ]
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file)
        return self

    @property
    def logging_dir(self) -> Path:
        return self.state_dir / 'logs'

    @property
    def http_cache_dir(self) -> Path:
        return self.cache_dir / '_http'

    @property
    def install_cache_dir(self) -> Path:
        return self.cache_dir / '_install'

    @property
    def plugins_cache_dir(self) -> Path:
        return self.cache_dir / 'plugins'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def profiles_config_dir(self) -> Path:
        return self.config_dir / 'profiles'

    @property
    def plugins_config_dir(self) -> Path:
        return self.config_dir / 'plugins'

    @property
    def profiles_state_dir(self) -> Path:
        return self.state_dir / 'profiles'

    @property
    def plugins_state_dir(self) -> Path:
        return self.state_dir / 'plugins'


@fauxfrozen
class _ProfileConfigStub:
    global_config: GlobalConfig
    profile: str = attrs.field(
        converter=str.strip,
        validator=_make_validate_min_length(1),
        metadata=FieldMetadata(env=True, store=True),
    )

    @property
    def config_dir(self) -> Path:
        return self.global_config.profiles_config_dir / self.profile

    @property
    def state_dir(self) -> Path:
        return self.global_config.profiles_state_dir / self.profile

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def db_file(self) -> Path:
        return self.config_dir / 'db.sqlite'


@fauxfrozen
class ProfileConfig(_ProfileConfigStub):
    addon_dir: Path = attrs.field(
        converter=_expand_path,
        validator=_validate_path_is_writable_dir,
        metadata=FieldMetadata(env=True, store=True),
    )
    game_flavour: Flavour = attrs.field(
        metadata=FieldMetadata(env=True, store=True),
    )

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, global_config: GlobalConfig, profile: str, *, env: bool = True) -> Self:
        return cls.from_values(
            {
                **read_config(cls, _ProfileConfigStub(global_config, profile).config_file),
                'global_config': global_config,
            },
            env=env,
        )

    def unstructure_for_display(self) -> str:
        return _unstructure_config_for_display(self)

    def ensure_dirs(self) -> Self:
        ensure_dirs(
            [
                self.config_dir,
                self.state_dir,
            ]
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file)
        return self

    def delete(self) -> None:
        trash((self.config_dir,))



from __future__ import annotations

from datetime import UTC, datetime

from .. import config_ctx
from .._utils.aio import run_in_thread
from ..catalogue.cataloguer import CatalogueEntry
from ..definitions import ChangelogFormat, Defn, SourceMetadata
from ..resolvers import BaseResolver, PkgCandidate
from ..results import PkgNonexistent
from ..wow_installations import Flavour

_ADDONS = {
    ('0', 'weakauras-companion'): False,
    ('1', 'weakauras-companion-autoupdate'): True,
}


class InstawowResolver(BaseResolver):
    metadata = SourceMetadata(
        id='instawow',
        name='instawow',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key=None,
    )

    async def resolve_one(self, defn: Defn, metadata: None):
        from instawow_wa_updater._config import PluginConfig
        from instawow_wa_updater._core import WaCompanionBuilder

        try:
            (id_, slug), requires_build = next(
                (p, v) for p, v in _ADDONS.items() if defn.alias in p
            )
        except StopIteration:
            raise PkgNonexistent from None

        builder_config = PluginConfig(config_ctx.config())
        builder = WaCompanionBuilder(builder_config)
        if requires_build:
            await run_in_thread(builder_config.ensure_dirs)()
            await builder.build()

        return PkgCandidate(
            id=id_,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=builder.build_paths.archive.as_uri(),
            date_published=datetime.now(UTC),
            version=await run_in_thread(builder.build_paths.version.read_text)(encoding='utf-8'),
            changelog_url=builder.build_paths.changelog.as_uri(),
        )

    @classmethod
    async def catalogue(cls):
        yield CatalogueEntry(
            source=cls.metadata.id,
            id='1',
            slug='weakauras-companion-autoupdate',
            name='WeakAuras Companion',
            url='https://github.com/layday/instawow',
            game_flavours=frozenset(Flavour),
            download_count=1,
            last_updated=datetime.now(UTC),
            folders=[
                frozenset({'WeakAurasCompanion'}),
            ],
        )



from __future__ import annotations

from pathlib import Path

import aiohttp
import aiohttp.web
import pytest

from instawow import config_ctx, pkg_management
from instawow.definitions import Defn, Strategy
from instawow.results import (
    InternalError,
    PkgAlreadyInstalled,
    PkgConflictsWithInstalled,
    PkgConflictsWithUnreconciled,
    PkgInstalled,
    PkgNonexistent,
    PkgNotInstalled,
    PkgRemoved,
    PkgSourceInvalid,
    PkgStrategiesUnsupported,
    PkgUpdated,
    PkgUpToDate,
)

from ._fixtures.http import ResponsesMockServer, Route

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


async def test_pinning_supported_pkg():
    defn = Defn('curse', 'molinari')

    install_result = (await pkg_management.install([defn], replace_folders=False))[defn]
    assert type(install_result) is PkgInstalled

    for new_defn in (defn.with_version(install_result.pkg.version), defn):
        pin_result = (await pkg_management.pin([new_defn]))[new_defn]
        assert type(pin_result) is PkgInstalled
        assert pin_result.pkg.options.version_eq is bool(new_defn.strategies[Strategy.VersionEq])
        assert install_result.pkg.version == pin_result.pkg.version


async def test_pinning_unsupported_pkg():
    molinari_defn = Defn('wowi', '13188')

    await pkg_management.install([molinari_defn], replace_folders=False)
    [installed_pkg] = pkg_management.get_pkgs([molinari_defn])
    assert installed_pkg is not None
    assert installed_pkg.options.version_eq is False

    result = (await pkg_management.pin([molinari_defn]))[molinari_defn]
    assert type(result) is PkgStrategiesUnsupported
    assert result.strategies == {Strategy.VersionEq}
    assert installed_pkg.options.version_eq is False


async def test_pinning_nonexistent_pkg():
    molinari_defn = Defn('curse', 'molinari')
    result = await pkg_management.pin([molinari_defn])
    assert type(result[molinari_defn]) is PkgNotInstalled


async def test_pinning_unsupported_nonexistent_pkg():
    molinari_defn = Defn('wowi', '13188')
    result = await pkg_management.pin([molinari_defn])
    assert type(result[molinari_defn]) is PkgStrategiesUnsupported


@pytest.mark.parametrize('exception', [ValueError('foo'), aiohttp.ClientError('bar')])
async def test_resolve_rewraps_exception_from_resolve(
    monkeypatch: pytest.MonkeyPatch,
    exception: Exception,
):
    async def resolve_one(defn, metadata):
        raise exception

    resolvers = config_ctx.resolvers()
    monkeypatch.setattr(resolvers['curse'], 'resolve_one', resolve_one)

    defn = Defn('curse', 'molinari')
    result = (await pkg_management.resolve([defn]))[defn]
    assert type(result) is InternalError
    assert str(result) == f'internal error: "{exception}"'


async def test_resolve_invalid_source():
    defn = Defn('bar', 'baz')
    results = await pkg_management.resolve([defn])
    assert type(results[defn]) is PkgSourceInvalid


async def test_resolve_plugin_hook_source():
    pytest.importorskip('instawow_test_plugin')
    defn = Defn('me', 'bar')
    results = await pkg_management.resolve([defn])
    assert type(results[defn]) is dict


async def test_install_can_replace_unreconciled_folders():
    molinari = config_ctx.config().addon_dir / 'Molinari'
    molinari.mkdir()

    defn = Defn('curse', 'molinari')

    result = await pkg_management.install([defn], replace_folders=False)
    assert type(result[defn]) is PkgConflictsWithUnreconciled
    assert not any(molinari.iterdir())

    result = await pkg_management.install([defn], replace_folders=True)
    assert type(result[defn]) is PkgInstalled
    assert any(molinari.iterdir())


async def test_install_cannot_replace_reconciled_folders():
    curse_defn = Defn('curse', 'molinari')
    wowi_defn = Defn('wowi', '13188-molinari')

    result = await pkg_management.install([curse_defn], replace_folders=False)
    assert type(result[curse_defn]) is PkgInstalled

    result = await pkg_management.install([wowi_defn], replace_folders=False)
    assert type(result[wowi_defn]) is PkgConflictsWithInstalled

    result = await pkg_management.install([wowi_defn], replace_folders=True)
    assert type(result[wowi_defn]) is PkgConflictsWithInstalled


async def test_install_recognises_renamed_pkg_from_id(
    monkeypatch: pytest.MonkeyPatch,
    iw_aresponses: ResponsesMockServer,
):
    iw_aresponses.add(
        Route(
            r'//api\.github\.com/repos/p3lim-wow/molinarifico/?.*',
            aiohttp.web.Response(status=404),
        )
    )

    old_defn = Defn('github', 'p3lim-wow/molinari')
    new_defn = Defn('github', 'p3lim-wow/molinarifico')

    result = await pkg_management.install([old_defn], replace_folders=False)
    assert type(result[old_defn]) is PkgInstalled

    result = await pkg_management.install([old_defn], replace_folders=False)
    assert type(result[old_defn]) is PkgAlreadyInstalled

    result = await pkg_management.install([new_defn], replace_folders=False)
    assert type(result[new_defn]) is PkgNonexistent

    old_resolve = pkg_management.resolve

    async def new_resolve(defns, with_deps=False):
        result = await old_resolve([old_defn])
        pkg_candidate = result[old_defn]
        assert type(pkg_candidate) is dict
        return {new_defn: pkg_candidate | {'slug': new_defn.alias}}

    monkeypatch.setattr(pkg_management, 'resolve', new_resolve)

    result = await pkg_management.install([new_defn], replace_folders=False)
    assert type(result[new_defn]) is PkgAlreadyInstalled


async def test_update_lifecycle_with_strategy_switch():
    defn = Defn('curse', 'molinari')
    versioned_defn = defn.with_version('100005.97-Release')

    result = (await pkg_management.install([defn], replace_folders=False))[defn]
    assert type(result) is PkgInstalled

    result = (await pkg_management.update([defn]))[defn]
    assert type(result) is PkgUpToDate
    assert result.is_pinned is False

    result = (await pkg_management.update([versioned_defn]))[versioned_defn]
    assert type(result) is PkgUpdated
    assert result.new_pkg.options.version_eq is True

    result = (await pkg_management.update([defn]))[defn]
    assert type(result) is PkgUpToDate
    assert result.is_pinned is True


async def test_update_reinstalls_corrupted_pkgs():
    defn = Defn('curse', 'molinari')

    results = await pkg_management.install([defn], replace_folders=False)

    result = results[defn]
    assert type(result) is PkgInstalled

    folders = [config_ctx.config().addon_dir / f.name for f in result.pkg.folders]

    first_folder = folders[0]
    first_folder.rename(first_folder.with_name('foo'))
    assert not all(f.is_dir() for f in folders)

    results = await pkg_management.update([defn])
    assert type(results[defn]) is PkgUpdated
    assert all(f.is_dir() for f in folders)


@pytest.mark.parametrize('keep_folders', [True, False])
async def test_deleting_and_retaining_folders_on_remove(
    keep_folders: bool,
):
    defn = Defn('curse', 'molinari')

    await pkg_management.install([defn], replace_folders=False)
    [pkg] = pkg_management.get_pkgs([defn])
    assert pkg

    folders = [config_ctx.config().addon_dir / f.name for f in pkg.folders]
    assert all(f.is_dir() for f in folders)

    results = await pkg_management.remove([defn], keep_folders=keep_folders)
    assert type(results[defn]) is PkgRemoved
    [pkg] = pkg_management.get_pkgs([defn])
    assert not pkg
    if keep_folders:
        assert all(f.is_dir() for f in folders)
    else:
        assert not any(f.is_dir() for f in folders)


@pytest.mark.parametrize('keep_folders', [True, False])
async def test_removing_pkg_with_missing_folders(
    keep_folders: bool,
):
    defn = Defn('curse', 'molinari')

    results = await pkg_management.install([defn], replace_folders=False)

    result = results[defn]
    assert type(result) is PkgInstalled

    folders = [config_ctx.config().addon_dir / f.name for f in result.pkg.folders]
    for folder in folders:
        folder.rename(folder.with_name(f'Not_{folder.name}'))
    assert not any(f.is_dir() for f in folders)

    results = await pkg_management.remove([defn], keep_folders=keep_folders)
    assert type(results[defn]) is PkgRemoved
    [pkg] = pkg_management.get_pkgs([defn])
    assert not pkg


async def test_replace_pkg():
    old_defn = Defn('curse', 'molinari')
    new_defn = Defn('github', 'p3lim-wow/molinari')

    results = await pkg_management.install([old_defn], replace_folders=False)
    assert type(results[old_defn]) is PkgInstalled

    results = await pkg_management.replace({old_defn: new_defn})
    assert len(results) == 2
    assert type(results[old_defn]) is PkgRemoved
    assert type(results[new_defn]) is PkgInstalled


async def test_get_changelog_from_empty_data_url():
    assert (await pkg_management.get_changelog('github', 'data:,')) == ''


async def test_get_changelog_from_url_encoded_data_url():
    assert (await pkg_management.get_changelog('github', 'data:,foo%20bar')) == 'foo bar'


async def test_get_malformed_changelog():
    with pytest.raises(ValueError, match='Unsupported URI with scheme'):
        await pkg_management.get_changelog('github', '')


async def test_get_changelog_from_file_uri(
    tmp_path: Path,
):
    changelog = tmp_path / 'changelog.txt'
    changelog.write_text('test')
    assert (await pkg_management.get_changelog('github', changelog.as_uri())) == 'test'


async def test_get_changelog_from_web_url():
    assert (
        await pkg_management.get_changelog(
            'curse',
            'https://api.curseforge.com/v1/mods/20338/files/3657564/changelog',
        )
    ).startswith('<h3>Changes in')



from __future__ import annotations

import json
import shutil
from functools import partial
from textwrap import dedent
from unittest import mock

import click.testing
import prompt_toolkit.input
import pytest

from instawow.cli import cli
from instawow.config import ProfileConfig

pytestmark = pytest.mark.usefixtures(
    '_iw_config_ctx', '_iw_web_client_ctx', '_iw_mock_pt_progress_bar', '_iw_mock_asyncio_run'
)


_runner = click.testing.CliRunner()
run = partial(_runner.invoke, cli, catch_exceptions=False)


def install_molinari():
    run('install curse:molinari')


def pretend_install_molinari(
    iw_profile_config: ProfileConfig,
):
    molinari = iw_profile_config.addon_dir / 'Molinari'
    molinari.mkdir()
    (molinari / 'Molinari.toc').write_text(
        """\
## X-Curse-Project-ID: 20338
## X-WoWI-ID: 13188
"""
    )


@pytest.mark.parametrize(
    'alias',
    [
        'curse:molinari',
        'wowi:13188-molinari',
        'tukui:tukui',
        'github:p3lim-wow/Molinari',
    ],
)
def test_valid_pkg_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').output.startswith(f'✓ {alias}\n  installed')
    assert run(f'install {alias}').output == f'✗ {alias}\n  package already installed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is up to date\n'
    assert run(f'remove {alias}').output == f'✓ {alias}\n  removed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['instawow:gargantuan-wigs'])
def test_nonexistent_pkg_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ {alias}\n  package does not exist\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('iw_global_config_values', [None], indirect=True)
@pytest.mark.parametrize('alias', ['curse:molinari'])
def test_disabled_source_lifecycle(
    alias: str,
):
    assert (
        run(f'install {alias}').output
        == f'✗ {alias}\n  package source is disabled: access token missing\n'
    )
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['foo:bar'])
def test_invalid_source_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ :{alias}\n  package source is invalid\n'
    assert run(f'update {alias}').output == f'✗ :{alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


def test_reconciled_folder_conflict_on_install():
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert run('install wowi:13188-molinari').output == (
        '✗ wowi:13188-molinari\n'
        '  package folders conflict with installed package Molinari\n'
        '    (curse:20338)\n'
    )


def test_unreconciled_folder_conflict_on_install(iw_profile_config):
    iw_profile_config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == "✗ curse:molinari\n  package folders conflict with 'Molinari'\n"
    )
    assert run('install --replace curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed'
    )


def test_keep_folders_on_remove(iw_profile_config):
    install_molinari()
    assert run('remove --keep-folders curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert iw_profile_config.addon_dir.joinpath('Molinari').is_dir()


def test_version_strategy_lifecycle():
    assert run('install curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed 100205.111-Release_5090686'
    )
    assert (
        run('install curse:molinari#version_eq=foo').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is up to date\n'
    assert run('update curse:molinari#version_eq=100005.97-Release').output == dedent(
        """\
        ✓ curse:molinari
          updated 100205.111-Release_5090686 to 100005.97-Release_4419396 with
            new strategies: version_eq='100005.97-Release_4419396'
        """
    )
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert run('install curse:molinari#version_eq=foo').output == dedent(
        """\
        ✗ curse:molinari
          no files are available for download
        """
    )
    assert (
        run('install curse:molinari#version_eq=100005.97-Release').output
        == '✓ curse:molinari\n  installed 100005.97-Release_4419396\n'
    )
    assert run('update').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('update curse:molinari#=').output == dedent(
        """\
        ✓ curse:molinari
          updated 100005.97-Release_4419396 to 100205.111-Release_5090686 with
            new strategies: version_eq=None
        """
    )
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


def test_install_options():
    assert run('install curse:molinari#any_release_type,any_flavour').output == dedent(
        """\
        ✓ curse:molinari
          installed 100205.111-Release_5090686
        """
    )


@pytest.mark.parametrize('step', [1, -1])
def test_install_order_is_respected(
    step: int,
):
    assert run(
        'install '
        + ' '.join(
            [
                'curse:molinari',
                'curse:molinari#version_eq=100005.97-Release',
            ][::step]
        )
    ).output == dedent(
        f"""\
        ✓ curse:molinari
          installed {'100205.111-Release_5090686' if step == 1 else '100005.97-Release_4419396'}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        """
    )


def test_install_invalid_defn():
    result = run('install foo')
    assert result.exit_code == 2
    assert result.output.endswith("Error: Invalid value for '[ADDONS]...': foo\n")


def test_install_dry_run():
    assert run('install --dry-run curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          would have installed 100205.111-Release_5090686
        """
    )


def test_debug_config(iw_profile_config):
    assert (
        run('debug config').output
        == json.dumps(iw_profile_config.unstructure_for_display(), indent=2) + '\n'
    )


def test_debug_sources():
    import cattrs.preconf.json

    from instawow.definitions import SourceMetadata

    json_converter = cattrs.preconf.json.make_converter()

    output = run('debug sources').output
    source_metadata = json_converter.loads(output, list[SourceMetadata])
    assert len(source_metadata) > 1


@pytest.mark.parametrize('command', ['configure', 'list'], ids=['explicit', 'implicit'])
def test_configure__create_new_profile(
    monkeypatch: pytest.MonkeyPatch,
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
    command: str,
):
    # In case there is a WoW installation in the test environment.
    monkeypatch.setattr('instawow.wow_installations.find_installations', lambda: iter([]))

    iw_pt_input.send_text(f'{iw_profile_config.addon_dir}\r\rY\r')
    assert run(f'-p foo {command}').output == (
        'Generating an access token for GitHub is recommended to avoid being\n'
        'rate limited.  You may only perform 60 requests an hour without an\n'
        'access token.\n'
        'Navigate to https://github.com/login/device and paste the code below:\n'
        '  WDJB-MJHT\n'
        'Waiting...\n'
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/foo/config.json"}\n'
    )


def test_configure__update_existing_profile_interactively(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    iw_pt_input.send_text('Y\r')
    assert run('configure global_config.auto_update_check').output == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/__default__/config.json"}\n'
    )


def test_configure__update_existing_profile_directly(
    iw_profile_config: ProfileConfig,
):
    assert run('configure global_config.auto_update_check=0').output == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/__default__/config.json"}\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__pkg_not_installed(
    options: str,
):
    assert (
        run(f'rollback {options} curse:molinari').output
        == '✗ curse:molinari\n  package is not installed\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__unsupported(
    options: str,
):
    assert run('install wowi:13188-molinari').exit_code == 0
    assert (
        run(f'rollback {options} wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  strategies are not valid for source: version_eq\n'
    )


def test_rollback__single_version():
    assert run('install curse:molinari').exit_code == 0
    assert (
        run('rollback curse:molinari').output == '✗ curse:molinari\n  cannot find older versions\n'
    )


def test_rollback__multiple_versions(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    assert run('install curse:molinari#version_eq=100005.97-Release').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    iw_pt_input.send_text('\r\r')
    assert run('rollback curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          updated 100205.111-Release_5090686 to 100005.97-Release_4419396 with
            new strategies: version_eq='100005.97-Release_4419396'
        """
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__rollback_multiple_versions(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    options: str,
):
    assert run('install curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari#version_eq=100005.97-Release').exit_code == 0
    iw_pt_input.send_text('\r\r')
    assert run(f'rollback {options} curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          updated 100005.97-Release_4419396 to 100205.111-Release_5090686 with
            new strategies: version_eq=None
        """
        if options == '--undo'
        else """\
        ✓ curse:molinari
          updated 100005.97-Release_4419396 to 100205.111-Release_5090686
        """
    )


def test_reconcile__list_unreconciled(
    iw_profile_config: ProfileConfig,
):
    pretend_install_molinari(iw_profile_config)
    assert run('reconcile --list-unreconciled').output == (
        'unreconciled\n'  # fmt: skip
        '------------\n'
        'Molinari    \n'
    )


def test_reconcile_leftovers(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_molinari(iw_profile_config)
    iw_pt_input.send_text('sss')  # Skip
    assert run('reconcile').output.endswith(
        'unreconciled\n'  # fmt: skip
        '------------\n'
        'Molinari    \n'
    )


def test_reconcile__auto_reconcile(iw_profile_config):
    pretend_install_molinari(iw_profile_config)
    assert run('reconcile --auto').output == dedent(
        """\
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_reconcile__abort_interactive_reconciliation(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_molinari(iw_profile_config)
    iw_pt_input.send_text('\x03')  # ^C
    assert run('reconcile').output.endswith('Aborted!\n')


def test_reconcile__complete_interactive_reconciliation(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_molinari(iw_profile_config)
    iw_pt_input.send_text('\ry')
    assert run('reconcile').output.endswith(
        dedent(
            """\
            ✓ github:p3lim-wow/molinari
              installed 100205.111-Release
            """
        )
    )


def test_reconcile__reconciliation_complete():
    assert run('reconcile').output == 'No add-ons left to reconcile.\n'


def test_rereconcile(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    install_molinari()
    iw_pt_input.send_text('\ry')
    assert run('rereconcile').output == dedent(
        """\
        ✓ curse:molinari
          removed
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_rereconcile_with_args(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    install_molinari()

    iw_pt_input.send_text('\ry')
    assert run('rereconcile foo').output == ''

    iw_pt_input.send_text('\ry')
    assert run('rereconcile molinari').output == dedent(
        """\
        ✓ curse:molinari
          removed
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_search__no_results():
    assert run('search ∅').output == 'No results found.\n'


def test_search__exit_without_selecting(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text('\r')  # enter
    assert run('search molinari').output == (
        'Nothing was selected; select add-ons with <space> and confirm by pressing <enter>.\n'
    )


def test_search__exit_after_selection(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \rn')  # space, enter, "n"
    assert run('search molinari').output == ''


def test_search__install_one(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \ry')  # space, enter, enter
    assert run('search molinari --source curse').output == dedent(
        """\
        ✓ curse:molinari
          installed 100205.111-Release_5090686
        """
    )


def test_search__install_multiple_conflicting(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \x1b[B \ry')  # space, arrow down, space, enter, enter
    assert run('search molinari').output == dedent(
        """\
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        ✗ wowi:13188
          package folders conflict with installed package Molinari
            (github:388670)
        """
    )


def test_changelog_output_installed_convert():
    install_molinari()
    output = (
        'curse:molinari:\n  Changes in 90200.82-Release:'
        if shutil.which('pandoc')
        else 'curse:molinari:\n  <h3>Changes in'
    )
    assert run('view-changelog curse:molinari').output.startswith(output)


def test_changelog_output_installed_no_convert():
    install_molinari()
    assert run('view-changelog --no-convert curse:molinari').output.startswith(
        'curse:molinari:\n  <h3>Changes in'
    )


def test_changelog_output_not_installed():
    assert run('view-changelog curse:molinari').output == ''


def test_changelog_output_argless():
    install_molinari()
    assert run('view-changelog').output.startswith('curse:molinari:')


def test_changelog_output_remote():
    assert run('view-changelog --remote curse:molinari').output.startswith('curse:molinari:')


@pytest.mark.parametrize(
    ('command', 'exit_code'),
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(
    monkeypatch: pytest.MonkeyPatch,
    command: str,
    exit_code: int,
):
    install_molinari()
    monkeypatch.setattr('instawow._utils.file.reveal_folder', mock.MagicMock())
    assert run(command).exit_code == exit_code


def test_can_list_with_substr_match():
    install_molinari()
    assert run('list mol').output == 'curse:molinari\n'
    assert run('list foo').output == ''
    assert run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export():
    install_molinari()
    output = run('list -f json').output
    assert json.loads(output)[0]['name'] == 'Molinari'


def test_show_version():
    from instawow._version import get_version

    assert run('--version').output == f'instawow, version {get_version()}\n'


def test_plugin_hook_command_can_be_invoked():
    pytest.importorskip('instawow_test_plugin')
    assert run('plugins foo').output == 'success!\n'


# Skip loading the `_iw_web_client_ctx` fixture
# to allow deleting the cache on Windows.
@pytest.mark.parametrize('_iw_web_client_ctx', [None])
def test_clear_cache(
    iw_profile_config: ProfileConfig,
):
    assert iw_profile_config.global_config.cache_dir.is_dir()
    assert run('cache clear').exit_code == 0
    assert not iw_profile_config.global_config.cache_dir.is_dir()



from __future__ import annotations

import json
import os
from collections.abc import Iterable, Mapping
from functools import lru_cache, partial
from pathlib import Path
from typing import Any, TypedDict, TypeVar

import attrs
import cattrs
import cattrs.gen
import cattrs.preconf.json

_T = TypeVar('_T')


_missing = object()


class FieldMetadata(TypedDict, total=False):
    env: bool
    preparse_from_env: bool
    store: bool


class UninitialisedConfigError(Exception):
    pass


def ensure_dirs(dirs: Iterable[Path]) -> None:
    for dir_ in dirs:
        dir_.mkdir(exist_ok=True, parents=True)


def make_config_converter() -> cattrs.Converter:
    converter = cattrs.Converter()
    cattrs.preconf.json.configure_converter(converter)
    converter.register_structure_hook(Path, lambda v, _: Path(v))
    converter.register_unstructure_hook(Path, str)
    return converter


@lru_cache(1)
def _make_write_converter():
    converter = make_config_converter()

    @converter.register_unstructure_hook_factory(attrs.has)
    def _(type_: type[_T]):
        return cattrs.gen.make_dict_unstructure_fn(
            type_,
            converter,
            **{  # pyright: ignore[reportArgumentType]  # See microsoft/pyright#5255
                f.name: cattrs.gen.override(omit=True)
                for f in attrs.fields(type_)
                if not f.metadata.get('store')
            },
        )

    return converter


def write_config(config: object, config_path: Path) -> None:
    converter = _make_write_converter()
    config_path.write_text(
        json.dumps(converter.unstructure(config), indent=2),
        encoding='utf-8',
    )


def read_config(config_cls: type, config_path: Path, missing_ok: bool = False) -> dict[str, Any]:
    try:
        values = json.loads(config_path.read_bytes())
        for field in attrs.fields(config_cls):
            if attrs.has(field.type):
                maybe_values = read_config(
                    field.type, config_path.with_stem(f'{config_path.stem}.{field.name}'), True
                )
                if maybe_values:
                    values |= {field.name: maybe_values}

    except FileNotFoundError:
        if missing_ok:
            return {}
        raise UninitialisedConfigError from None
    else:
        return values


def _compute_var(field: attrs.Attribute[object], default: object):
    env = field.metadata.get('env')
    if not env:
        return default

    value = os.environ.get(f'{"instawow" if env is True else env}_{field.name}'.upper())
    if value is None:
        return default
    elif field.metadata.get('preparse_from_env'):
        return json.loads(value)
    else:
        return value


def read_env_vars(config_cls: type, values: Mapping[str, object]) -> dict[str, object]:
    return {
        f.name: v
        for f in attrs.fields(config_cls)
        for v in (_compute_var(f, values.get(f.name, _missing)),)
        if v is not _missing
    }


def _make_attrs_instance_hook_factory(converter: cattrs.Converter, type_: type[Any]):
    "Allow passing in a structured attrs instance to ``structure``."
    structure = converter.gen_structure_attrs_fromdict(type_)

    def structure_wrapper(value: Mapping[str, Any], type_: type[Any]):
        return value if isinstance(value, type_) else structure(value, type_)

    return structure_wrapper


config_converter = make_config_converter()
config_converter.register_structure_hook_factory(
    attrs.has, partial(_make_attrs_instance_hook_factory, config_converter)
)



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

import attrs

_T = TypeVar('_T')

_Validator: TypeAlias = Callable[[object, 'attrs.Attribute[_T]', _T], None]


fauxfrozen = attrs.frozen if TYPE_CHECKING else partial(attrs.define, unsafe_hash=True)


def enrich_validator_exc(validator: _Validator[_T]) -> _Validator[_T]:
    "Pretend validation error originates from cattrs for uniformity with structural errors."

    import cattrs

    def wrapper(model: object, attr: attrs.Attribute[_T], value: _T):
        try:
            validator(model, attr, value)
        except BaseException as exc:
            note = f'Structuring class {model.__class__.__name__} @ attribute {attr.name}'
            exc.add_note(cattrs.AttributeValidationNote(note, attr.name, attr.type))
            raise

    return wrapper


@fauxfrozen
class EvolveIdent:
    value: Mapping[Any, Any]


def evolve(attrs_instance: _T, changes: Mapping[str, Any | EvolveIdent]) -> _T:
    return attrs.evolve(
        attrs_instance,
        **{
            k: evolve(
                getattr(attrs_instance, k),
                v,  # pyright: ignore[reportUnknownArgumentType]
            )
            if isinstance(v, Mapping)
            else v.value
            if isinstance(v, EvolveIdent)
            else v
            for k, v in changes.items()
        },
    )



from __future__ import annotations

import asyncio
import contextvars
from collections.abc import AsyncGenerator, Awaitable, Callable, Iterator, Mapping
from contextlib import contextmanager
from functools import partial
from itertools import count
from typing import Any, Generic, Literal, LiteralString, Never, NotRequired, TypeAlias

from typing_extensions import TypedDict, TypeVar

_T = TypeVar('_T')

_TProgressType = TypeVar('_TProgressType', bound=LiteralString)
_TProgressUnit = TypeVar('_TProgressUnit', bound=LiteralString | None, default=None)


class Progress(TypedDict, Generic[_TProgressType, _TProgressUnit]):
    type_: _TProgressType
    unit: NotRequired[_TProgressUnit]
    label: NotRequired[str]
    current: int
    total: int | None


class _GenericProgress(Progress[Literal['generic']]):
    pass


class _DownloadProgress(Progress[Literal['download'], Literal['bytes']]):
    pass


make_generic_progress = partial(_GenericProgress, type_='generic', current=0, total=0)
make_download_progress = partial(
    _DownloadProgress, type_='download', unit='bytes', current=0, total=0
)


_TProgress = TypeVar('_TProgress', bound=Progress[Any, Any], default=Never)
ReadOnlyProgressGroup: TypeAlias = Mapping[int, _DownloadProgress | _GenericProgress | _TProgress]

_progress_notifiers_var = contextvars.ContextVar[
    frozenset[Callable[[int, Progress[Any, Any] | Literal['unset']], None]]
]('_progress_notifiers_var', default=frozenset())

_progress_id_gen = count()


def get_next_progress_id() -> int:
    return next(_progress_id_gen)


class make_progress_receiver(Generic[_TProgress]):
    "Observe progress within the current context."

    @contextmanager
    def __new__(
        cls,
    ) -> Iterator[
        tuple[
            Callable[[], ReadOnlyProgressGroup[_TProgress]],
            Callable[[], AsyncGenerator[ReadOnlyProgressGroup[_TProgress]]],
        ]
    ]:
        asyncio.get_running_loop()  # Raise if constructed outside async context.

        emit_event = asyncio.Event()

        progress_group: ReadOnlyProgressGroup[_TProgress] = {}

        def waken(progress_id: int, progress: Progress[Any, Any] | Literal['unset']):
            if progress == 'unset':
                try:
                    del progress_group[progress_id]
                except KeyError:
                    pass
            else:
                progress_group[progress_id] = progress

            emit_event.set()

        _progress_notifiers_var.set(_progress_notifiers_var.get() | {waken})

        try:

            def get_once():
                return progress_group

            async def make_iter():
                while True:
                    await emit_event.wait()
                    yield progress_group
                    emit_event.clear()

            yield (get_once, make_iter)

        finally:
            _progress_notifiers_var.set(_progress_notifiers_var.get() - {waken})


def update_progress(progress_id: int, progress: Progress[Any, Any] | Literal['unset']) -> None:
    "Trigger a manual progress update."
    for notify in _progress_notifiers_var.get():
        notify(progress_id, progress)


def make_incrementing_progress_tracker(
    total: int, label: str
) -> Callable[[Awaitable[_T]], Awaitable[_T]]:
    "Track the progress a finite-length collection of awaitables."
    if total < 1:

        def track_ident(awaitable: Awaitable[_T]):
            return awaitable

        return track_ident

    progress_id = get_next_progress_id()
    progress = make_generic_progress(total=total, label=label)

    def track(awaitable: Awaitable[_T]):
        future = asyncio.ensure_future(awaitable)

        @future.add_done_callback
        def _(task: asyncio.Task[_T]):
            progress['current'] += 1
            update_progress(
                progress_id,
                'unset' if progress['current'] == progress['total'] else progress,
            )

        return future

    update_progress(progress_id, progress)
    return track
